from __future__ import absolute_import, unicode_literals


# CrossRef publisher -> DOI prefixes:  http://www.crossref.org/06members/50go-live.html
# independent CSV map of same (3 years ago): https://gist.github.com/hubgit/5974843
# a short list of major publishers: https://webhome.weizmann.ac.il/home/comartin/doi.html

HOSTNAME_TO_DOI_PREFIX_MAP = {'cancerres.aacrjournals.org': '10.1158',  #TODO: *.aacrjournals.org
                              'jcs.biologists.org': '10.1242',
                              'fasebj.org': '10.1096',
                              'ajcn.nutrition.org': '10.3945',
                              'mcponline.org': '10.1074',
                              'karger.com': '10.1159',
                              '*.biomedcentral.com': '10.1186',
                              'haematologica.org': '10.3324',
                              'genesdev.cshlp.org': '10.1101',
                              '*.bmj.com': '10.1136',
                              'jbc.org': '10.1074',
                              '*.oxfordjournals.org': '10.1093',
                              'eje-online.org': '10.1530',
                              'jimmunol.org': '10.4049',
                              '*.asm.org': '10.1128',
                              'elifesciences.org': '10.7554',
                              }




from __future__ import absolute_import, unicode_literals

# Below: Journals with really annoying paywalls guarding their precious
# secrets.
schattauer_journals = [
    'Thromb Haemost',
]

# Royal Society of Chemistry
RSC_journals = ['Nat Prod Rep']

# thieme journals so far don't seem to have any open access content.
# example links to article page: https://www.thieme-connect.com/DOI/DOI?10.1055/s-0028-1085467
#           https://www.thieme-connect.com/DOI/DOI?10.1055/s-2007-1004566
thieme_journals = ['Neuropediatrics', 'Semin Vasc Med', 'Exp Clin Endocrinol Diabetes',
                   'Int J Sports Med', 'Horm Metab Res', 'Drug Res (Stuttg)',
                   ]

weird_paywall_publishers = ['J Ment Health Policy Econ']



from __future__ import absolute_import, unicode_literals

# Spandidos: volume/issue/firstpage AND a journal abbreviation. Fancy.
spandidos_format = 'http://www.spandidos-publications.com/{ja}/{a.volume}/{a.issue}/{a.first_page}/download'
spandidos_journals = {
    'Int J Oncol': {'ja': 'ijo'},
    'Int J Mol Med': {'ja': 'ijmm'},
    'Oncol Lett': {'ja': 'ol'},
    'Oncol Rep': {'ja': 'or'},
    'Mol Med Rep': {'ja': 'mmr'},
    'Biomed Rep': {'ja': 'br'},
    'Exp Ther Med': {'ja': 'etm'},
    'Mol Clin Oncol': {'ja': 'mco'},
}



from __future__ import absolute_import, unicode_literals

# simple formats are used for URLs that can be deduced from PubMedArticle XML
#
#       !ACHTUNG!  informa has been known to block IPs for the capital offense of
#                  having "More than 25 sessions created in 5 minutes"
#

doi_templates = {
    'acs': 'http://pubs.acs.org/doi/pdf/{a.doi}',
    'akademii': 'http://www.akademiai.com/content/{a.pii}/fulltext.pdf',
    'ats': 'http://www.atsjournals.org/doi/pdf/{a.doi}',
    'futuremed': 'http://www.futuremedicine.com/doi/pdf/{a.doi}',
    'informa': 'http://informahealthcare.com/doi/pdf/{a.doi}',
    'lancet': 'http://www.thelancet.com/pdfs/journals/{ja}/PII{a.pii}.pdf',
    'liebert': 'http://online.liebertpub.com/doi/pdf/{a.doi}',
    'plos': 'http://www.plosone.org/article/fetchObject.action?uri=info:doi/{a.doi}&representation=PDF',
    'taylor_francis': 'http://www.tandfonline.com/doi/pdf/{a.doi}',
    'wiley': 'http://onlinelibrary.wiley.com/doi/{a.doi}/pdf',
    'jci': 'http://www.jci.org/articles/view/{a.pii}/pdf',
}

simple_formats_doi = {
    'Acta Oncol': doi_templates['informa'],
    'Ann Hum Biol': doi_templates['informa'],
    'Hemoglobin': doi_templates['informa'],
    'J Matern Fetal Neonatal Med': doi_templates['informa'],
    'Ophthalmic Genet': doi_templates['informa'],
    'Platelets': doi_templates['informa'],
    'Ren Fail': doi_templates['informa'],
    'Scand J Rheumatol': doi_templates['informa'],
    'Scand J Urol Nephrol': doi_templates['informa'],
    'Xenobiotica': doi_templates['informa'],

    'Am J Public Health': 'http://ajph.aphapublications.org/doi/pdf/{a.doi}',
    'Am J Respir Cell Mol Biol': doi_templates['ats'],
    'Am J Respir Crit Care Med': doi_templates['ats'],

    'Anal Chem': doi_templates['acs'],
    'ACS Appl Mater': doi_templates['acs'],
    'ACS Chem Biol': doi_templates['acs'],
    'ACS Nano': doi_templates['acs'],
    'Biochemistry': doi_templates['acs'],
    'Chem Res Toxicol': doi_templates['acs'],
    'Inorg Chem': doi_templates['acs'],
    'J Agric Food Chem': doi_templates['acs'],
    'J Am Chem Soc': doi_templates['acs'],
    'J Comb Chem': doi_templates['acs'],
    'J Med Chem': doi_templates['acs'],
    'J Phys Chem A': doi_templates['acs'],
    'Langmuir': doi_templates['acs'],
    'Nano Lett': doi_templates['acs'],

    'BMJ Open Gastroenterol': 'http://bmjopengastro.bmj.com/doi/pdf/{a.doi}',
    'Microbiol Spectr': 'http://www.asmscience.org/content/journal/microbiolspec/{a.doi}', #10.1128/microbiolspec.VMBF-0028-2015

    # http://www.bioone.org/action/showPublications?type=byAlphabet
    # TODO: 'http://www.bioone.org/doi/pdf/{a.doi}',

    'AIDS Res Hum Retroviruses': doi_templates['liebert'],
    'Antioxid Redox Signal': doi_templates['liebert'],
    'Child Obes': doi_templates['liebert'],
    'DNA Cell Biol': doi_templates['liebert'],
    'Genet Test': doi_templates['liebert'],
    'Genet Test Mol Biomarkers': doi_templates['liebert'],
    'Thyroid': doi_templates['liebert'],
    'Vector Borne Zoonotic Dis': doi_templates['liebert'],

    # TODO: the rest of futuremed journals. see http://www.futuremedicine.com/
    'Pharmacogenomics': doi_templates['futuremed'],

    'Acta Orthop Scand': doi_templates['taylor_francis'],
    'Acta Orthop Scand Suppl': doi_templates['taylor_francis'],
    'Acta Otolaryngol': doi_templates['taylor_francis'],
    'Acta Otolaryngol Suppl': doi_templates['taylor_francis'],
    'Acta Otorhinolaryngol Belg': doi_templates['taylor_francis'],
    'Autophagy': doi_templates['taylor_francis'],
    'Biosci Biotechnol Biochem': doi_templates['taylor_francis'],
    'Cancer Biol Ther': doi_templates['taylor_francis'],
    'Cell Cycle': doi_templates['taylor_francis'],
    'Drug Metab Rev': doi_templates['taylor_francis'],
    'Environ Technol': doi_templates['taylor_francis'],
    'Health Commun': doi_templates['taylor_francis'],
    'J Biomol Struct Dyn': doi_templates['taylor_francis'],
    'J Pers Assess': doi_templates['taylor_francis'],
    'Pediatr Hematol Oncol': doi_templates['taylor_francis'],
    'Worm': doi_templates['taylor_francis'],

    'Endocrinology': 'http://press.endocrine.org/doi/pdf/{a.doi}',
    'Endocr Rev': 'http://press.endocrine.org/doi/pdf/{a.doi}',
    'Mol Endocrinol': 'http://press.endocrine.org/doi/pdf/{a.doi}',
    'J Periodontol': 'http://www.joponline.org/doi/pdf/{a.doi}',

    'PLoS Biol': doi_templates['plos'],
    'PLoS Comput Biol': doi_templates['plos'],
    'PLoS Genet': doi_templates['plos'],
    'PLoS Med': doi_templates['plos'],
    'PLoS ONE': doi_templates['plos'],
    'PLoS Pathog': doi_templates['plos'],
    'N Engl J Med':  'http://www.nejm.org/doi/pdf/{a.doi}',
}



from __future__ import absolute_import, print_function, unicode_literals

__author__ = 'nthmost'

import six

#py3k / py2k compatibility
if six.PY2:
    from urlparse import urlsplit, urljoin
else:
    from urllib.parse import urlsplit, urljoin

import requests
from lxml.html import HTMLParser
from lxml import etree

from ..dx_doi import DxDOI, DX_DOI_URL
from ..pubmedarticle import square_voliss_data_for_pma
from ..exceptions import AccessDenied, NoPDFLink, BadDOI, DxDOIError
from ..text_mining import find_doi_in_string
from ..utils import remove_chars

from .journals import *

OK_STATUS_CODES = (200, 301, 302, 307)

#TODO: make configurable (somehow...)
AAAS_USERNAME = 'nthmost'
AAAS_PASSWORD = '434264'

dx_doi_engine = None
def _start_dx_doi_engine():
    global dx_doi_engine
    if dx_doi_engine is None:
        dx_doi_engine = DxDOI()

def the_doi_2step(doi):
    '''Given a doi, uses DxDOI lookup engine to source the publisher's 
            article URL for this doi.

    Args:
        doi (str): a Digital Object Identifier

    Returns:
        url (str): link to publisher's website as returned by dx.doi.org

    Raises:
        NoPDFLink if dx.doi.org lookup failed (see error.message)
    '''
    _start_dx_doi_engine()
    try:
        return dx_doi_engine.resolve(doi)
    except (BadDOI, DxDOIError) as error:
        raise NoPDFLink('%r' % error)

def standardize_journal_name(journal_name):
    '''Returns a "standardized" journal name with periods stripped out.'''
    return remove_chars(journal_name, '.')

def verify_pdf_url(pdfurl, publisher_name=''):
    res = requests.get(pdfurl)
    if res.status_code==401:
        raise NoPDFLink('DENIED: %s url (%s) requires login.' % (publisher_name, pdfurl))

    if not res.ok:
        raise NoPDFLink('TXERROR: %i status returned from %s url (%s)' % (res.status_code, 
                            publisher_name, pdfurl))

    if res.status_code in OK_STATUS_CODES and res.headers['content-type'].find('pdf') > -1:
        return pdfurl
    else:
        raise NoPDFLink('DENIED: %s url (%s) did not result in a PDF' % (publisher_name, pdfurl))

def rectify_pma_for_vip_links(pma):
    '''takes a PubMedArticle object and "squares" the volume/issue/page info (sometimes there
    are weird characters in it, or sometimes the issue number is packed into the volume field,
    stuff like that).

    If volume, issue, and page data are all represented, return the PubMedArticle (possibly
    modified).

    If missing data, raise NoPDFLink('MISSING: vip...') 
    '''
    pma = square_voliss_data_for_pma(pma)
    if pma.volume and pma.first_page and pma.issue:
        return pma
    raise NoPDFLink('MISSING: vip (volume, issue, and/or first_page missing from PubMedArticle)')

def the_doi_slide(pma, verify=True):
    '''Dance of the miscellaneous journals that use DOI in their URL construction.

         :param: pma (PubMedArticle object)
         :param: verify (bool) [default: True]
         :return: url (string)
         :raises: AccessDenied, NoPDFLink
    '''
    jrnl = standardize_journal_name(pma.journal) 
    if pma.doi:
        url = simple_formats_doi[jrnl].format(a=pma)
    else:
        raise NoPDFLink('MISSING: doi (lookup failed)')

    if verify:
        verify_pdf_url(url)
    return url

def the_pmid_pogo(pma, verify=True):
    '''Dance of the miscellaneous journals that use PMID in their URL construction.

         :param: pma (PubMedArticle object)
         :param: verify (bool) [default: True]
         :return: url (string)
         :raises: AccessDenied, NoPDFLink
    '''
    jrnl = standardize_journal_name(pma.journal)
    url = simple_formats_pmid[jrnl].format(pmid=pma.pmid)

    if verify:
        verify_pdf_url(url)
    return url

def the_vip_shake(pma, verify=True):
    '''Dance of the miscellaneous journals that use volume-issue-page in their
        URL construction

         :param: pma (PubMedArticle object)
         :param: verify (bool) [default: True]
         :return: url (string)
         :raises: AccessDenied, NoPDFLink
    '''
    jrnl = standardize_journal_name(pma.journal)
    pma = rectify_pma_for_vip_links(pma)  #raises NoPDFLink if missing data.
    url = vip_format.format(host=vip_journals[jrnl]['host'], a=pma)

    if verify:
        verify_pdf_url(url)
    return url

def the_vip_nonstandard_shake(pma, verify=True):
    '''Dance of the miscellaneous journals that use volume-issue-page in their
        URL construction (but are a little different).

         :param: pma (PubMedArticle object)
         :param: verify (bool) [default: True]
         :return: url (string)
         :raises: AccessDenied, NoPDFLink
    '''
    jrnl = standardize_journal_name(pma.journal)
    pma = rectify_pma_for_vip_links(pma)  #raises NoPDFLink if missing data.
    url = vip_journals_nonstandard[jrnl].format(a=pma)

    if verify:
        verify_pdf_url(url)
    return url
 

def the_pii_polka(pma, verify=True):
    '''Dance of the miscellaneous journals that use a PII in their URL construction
        in their URL construction.

         :param: pma (PubMedArticle object)
         :param: verify (bool) [default: True]
         :return: url (string)
         :raises: AccessDenied, NoPDFLink
    '''
    jrnl = standardize_journal_name(pma.journal)
    if pma.pii:
        url = simple_formats_pii[jrnl].format(a=pma)
    else:
        raise NoPDFLink('MISSING: pii missing from PubMedArticle XML (pii format)')

    if url:
        res = requests.get(url)
        if res.text.find('Access Denial') > -1:
            raise AccessDenied('DENIED: Access Denied by ScienceDirect (%s)' % url)

    if verify:
        verify_pdf_url(url)
    return url


def the_jci_jig(pma, verify=True):
    '''Dance of the Journal of Clinical Investigation, which should be largely free.

         :param: pma (PubMedArticle object)
         :param: verify (bool) [default: True]
         :return: url (string)
         :raises: AccessDenied, NoPDFLink
    '''
    # approach: pii with dx.doi.org fallback to get pdf "view" page; scrape pdf download link.
    if pma.pii:
        starturl = doi_templates['jci'].format(a=pma)
    elif pma.doi:
        starturl = the_doi_2step(pma.doi)
        starturl = starturl + '/pdf'
    else:
        raise NoPDFLink('MISSING: pii, doi (doi lookup failed)')

    # Iter 1: do this until we see it stop working. (Iter 2: scrape download link from page.)
    url = starturl.replace('/pdf', '/version/1/pdf/render')
    if verify:
        verify_pdf_url(url, 'JCI')
    return url

def the_najms_mazurka(pma, verify=True):
    '''Dance of the North Am J Med Sci, which should be largely free.

         :param: pma (PubMedArticle object)
         :param: verify (bool) [default: True]
         :return: url (string)
         :raises: AccessDenied, NoPDFLink
    '''
    #PDF link looks like this:
    #http://www.najms.org/downloadpdf.asp?issn=1947-2714;year=2015;volume=7;issue=6;spage=291;epage=294;aulast=Thawabi;type=2

    url_tmpl = 'http://www.najms.org/downloadpdf.asp?issn={issn};year={a.year};volume={a.volume};issue={a.issue};spage={a.first_page};epage={a.last_page};aulast={author1_lastname};type=2'
    if pma.doi:
        #starturl = the_doi_2step(pma.doi)
        starturl = url_tmpl.format(a=pma, author1_lastname=pma.author1_last_fm.split(' ')[0], issn=pma.doi.split('/')[1].split('.')[0])
    else:
        raise NoPDFLink('MISSING: pii, doi (doi lookup failed)')

    url = ''
    response = requests.get(starturl)
    if response.ok:
        body = etree.fromstring(response.content, parser=HTMLParser()).find('body')
        href = body.findall('table/tr/td/p/a')[0].get('href')
        if href:
            url = urljoin('http://www.najms.org', href)
        else:
            raise NoPDFLink('TXERROR: NAJMS did not provide PDF link (or could not be parsed from page).')
    else:
        raise NoPDFLink('TXERROR: response from NAJMS website was %i' % response.status_code)

    if verify:
        verify_pdf_url(url, 'NAJMS')
    return url

def the_sciencedirect_disco(pma, verify=True):
    '''Note: verify=True required to find link.  Parameter supplied only for unity 
       with other dances.

         :param: pma (PubMedArticle object)
         :param: verify (bool) [default: True]
         :return: url (string)
         :raises: AccessDenied, NoPDFLink
    '''
    #we're looking for a url that looks like this:
    #http://www.sciencedirect.com/science/article/pii/S0022283601953379/pdfft?md5=07db9e1b612f64ea74872842e34316a5&pid=1-s2.0-S0022283601953379-main.pdf

    starturl = None
    if pma.pii:
        starturl = sciencedirect_url.format(piit=remove_chars(pma.pii, '-()'))
    elif pma.doi:
        starturl = the_doi_2step(pma.doi)

    if starturl == None:
        raise NoPDFLink('MISSING: pii, doi (doi lookup failed)')

    try:
        res = requests.get(starturl)
    except requests.exceptions.TooManyRedirects:
        raise NoPDFLink('TXERROR: ScienceDirect TooManyRedirects: cannot reach %s via %s' %
                        (pma.journal, starturl))

    tree = etree.fromstring(res.content, HTMLParser())
    try:
        div = tree.cssselect('div.icon_pdf')[0]
    except IndexError:
        raise NoPDFLink('DENIED: ScienceDirect did not provide pdf link (probably paywalled)')
    url = div.cssselect('a')[0].get('href')
    if url.find('.pdf') > -1:
        return url
    else:
        # give up, it's probably a "shopping cart" link.
        raise NoPDFLink('DENIED: ScienceDirect did not provide pdf link (probably paywalled)')

def the_biomed_calypso(pma, verify=False):
    '''Note: verification turned off by default because BMC is an all-open-access publisher.

       (You may still like to use verify=True to make sure it's a valid link.)

         :param: pma (PubMedArticle object)
         :param: verify (bool) [default: False]
         :return: url (string)
         :raises: AccessDenied, NoPDFLink
    '''
    baseid = pma.doi if pma.doi else pma.pii
    if baseid:
        article_id = baseid.split('/')[1]
    else:
        raise NoPDFLink('MISSING: doi needed for BMC article')
    url = BMC_format.format(aid=article_id)
    if verify:
        verify_pdf_url(url, 'BMC')
    return url


def the_scielo_chula(pma, verify=True):
    '''SciELO: The Scientific Electronic Library Online

    SciELO is an electronic library covering a selected collection of Brazilian
    scientific journals.

    Examples:
        23657305: pii = S0004-28032013000100035
            http://www.scielo.br/scielo.php?script=sci_arttext&pid=S0004-28032013000100035
            http://www.scielo.br/pdf/ag/v50n1/0004-2803-ag-50-01-35.pdf
    '''
    page_text = None
    baseurl_pii = 'http://www.scielo.br/scielo.php?script=sci_arttext&pid=%s'
    if pma.pii:
        response = requests.get(baseurl_pii % pma.pii)
        if response.ok:
            page_text = response.content

    if page_text is None:
        if pma.doi:
            response = requests.get(the_doi_2step(pma.doi))
            if response.ok:
                page_text = response.content
        else:
            raise NoPDFLink('MISSING: pii or doi needed for SciELO lookup.')

    if page_text:
        pdf_url = None
        head = etree.fromstring(page_text, HTMLParser()).getchildren()[0]
        for elem in head.findall('meta'):
            if elem.get('name') == 'citation_pdf_url':
                pdf_url = elem.get('content')

        if pdf_url:
            if verify:
                verify_pdf_url(pdf_url)
            return pdf_url
        else:
            #TODO: some other fallback manoeuvre?
            raise NoPDFLink('TXERROR: SciELO article page lacks PDF url. (See %s)' % response.url)

    else:
        raise NoPDFLink('TXERROR: SciELO page load responded with not-ok status: %i' % response.status_code)


def the_aaas_tango(pma, verify=True):
    '''Note: "verify" param recommended here (page navigation usually necessary).
    
         :param: pma (PubMedArticle object)
         :param: verify (bool) [default: True]
         :return: url (string)
         :raises: AccessDenied, NoPDFLink
    '''
    #try:
    #    pma = rectify_pma_for_vip_links(pma)
    #    pdfurl = aaas_format.format(ja=aaas_journals[pma.journal]['ja'], a=pma)
    #except NoPDFLink:
        # try the pmid-based approach
    baseurl = 'http://www.sciencemag.org/cgi/pmidlookup?view=long&pmid=%s' % pma.pmid
    res = requests.get(baseurl)
    pdfurl = res.url.replace('.long', '.full') + '.pdf'

    if not verify:
        return pdfurl

    response = requests.get(pdfurl)
    if response.status_code == 200 and response.headers['content-type'].find('pdf') > -1:
        return response.url

    elif response.status_code == 200 and response.headers['content-type'].find('html') > -1:
        tree = etree.fromstring(response.content, HTMLParser())

        if not tree.find('head/title').text.find('Sign In') > -1:
            raise NoPDFLink('TXERROR: AAAS returned unexpected HTML response for url %s' % (pdfurl))
        else:
            # some items are acquirable via free account registration... but let's not mess with this just yet.
            raise NoPDFLink('DENIED: AAAS paper subscription-only or requires site registration (url: %s)' % pdfurl)
    
        form = tree.cssselect('form')[0]
        fbi = form.fields.get('form_build_id')

        baseurl = urlsplit(response.url)
        post_url = baseurl.scheme + '://' + baseurl.hostname + form.action

        payload = {'pass': AAAS_PASSWORD, 'name': AAAS_USERNAME,
                   'form_build_id': fbi, 'remember_me': 1}
        print("SUBMITTING TO AAAS")
        print(payload)

        response = requests.post(post_url, data=payload)
        if response.status_code == 403:
            return AccessDenied('DENIED: AAAS subscription-only paper (url: %s)')
        elif response.headers['content-type'].find('pdf') > -1:
            return response.url
        elif response.headers['content-type'].find('html') > -1:
            #if response.content.find('access-denied') > -1:
            raise NoPDFLink('DENIED: AAAS subscription-only paper (url: %s)' % pdfurl)
    else:
        raise NoPDFLink('TXERROR: AAAS returned %s for url %s' % (response.status_code, pdfurl))

def the_jama_dance(pma, verify=True):
    '''  :param: pma (PubMedArticle object)
         :param: verify (bool) [default: True]
         :return: url (string)
         :raises: AccessDenied, NoPDFLink
    '''
    if not pma.doi:
        raise NoPDFLink('MISSING: doi needed for JAMA article.')

    baseurl = the_doi_2step(pma.doi)
    res = requests.get(baseurl)
    parser = HTMLParser()
    tree = etree.fromstring(res.content, parser)
    # we're looking for a meta tag like this:
    # <meta name="citation_pdf_url" content="http://archneur.jamanetwork.com/data/Journals/NEUR/13776/NOC40008.pdf" />
    for item in tree.findall('head/meta'):
        if item.get('name') == 'citation_pdf_url':
            pdfurl = item.get('content')
        else:
            raise NoPDFLink('DENIED: JAMA did not provide PDF link in (%s).' % baseurl)
    if verify:
        #TODO: form navigation
        verify_pdf_url(pdfurl, 'JAMA')
    return pdfurl

def the_jstage_dive(pma, verify=True):
    '''  :param: pma (PubMedArticle object)
         :param: verify (bool) [default: True]
         :return: url (string)
         :raises: AccessDenied, NoPDFLink
    '''
    if pma.doi:
        url = the_doi_2step(pma.doi)
        res = requests.get(url)
        if res.url.find('jstage') > -1:
            url = res.url.replace('_article', '_pdf')
            pdfpos = url.find('_pdf')
            # remove everything after the "_pdf" part
            url = url[:pdfpos+4]
    else:
        raise NoPDFLink('MISSING: doi for dx.doi.org lookup to get jstage link')

    if verify:
        verify_pdf_url(url, 'jstage')
    return url


def the_wiley_shuffle(pma, verify=True):
    '''Returns a PDF link to an article from a Wiley-published journal.

    Note: Wiley sometimes buries PDF links in HTML pages we have to parse first.
    Turning off verification (verify=False) may return only a superficial link.
         
         :param: pma (PubMedArticle object)
         :param: verify (bool) [default: True]
         :return: url (string)
         :raises: AccessDenied, NoPDFLink
    '''
    url = doi_templates['wiley'].format(a=pma)
    if not verify:
        return url

    # wiley sometimes buries PDF links in HTML pages we have to parse.
    res = requests.get(url)
    if res.headers['content-type'].find('html') > -1:
        if 'ACCESS DENIED' in res.text:
            raise AccessDenied('DENIED: Wiley E Publisher says no to %s' % res.url)

        tree = etree.fromstring(res.content, HTMLParser())
        if tree.find('head/title').text.find('Not Found') > -1:
            raise NoPDFLink('TXERROR: Wiley says File Not found (%s)' % res.url)
        elif tree.find('head/title').text.find('Maintenance') > -1:
            raise NoPDFLink('TXERROR: Wiley site under scheduled maintenance -- try again later (url was %s).' % res.url)
        iframe = tree.find('body/div/iframe')

        try:
            url = iframe.get('src')
        except AttributeError:
            # no iframe, give up (probably asking for a login at this point)
            raise AccessDenied('DENIED: Wiley E Publisher says no to %s' % res.url)
        verify_pdf_url(url, 'Wiley')
        return url

    elif res.headers['content-type'].find('pdf' > -1):
        return res.url


def the_lancet_tango(pma, verify=True):
    '''  :param: pma (PubMedArticle object)
         :param: verify (bool) [default: True]
         :return: url (string)
         :raises: AccessDenied, NoPDFLink
    '''
    url_template = doi_templates['lancet']
    jrnl = standardize_journal_name(pma.journal)

    if not pma.pii and pma.doi:
        pma.pii = pma.doi.split('/')[1]
    elif not pma.pii and not pma.doi:
        raise NoPDFLink('MISSING: pii, doi (doi lookup failed)')

    # if we got here, we should have a pii.
    url = url_template.format(a=pma, ja=lancet_journals[jrnl]['ja'])
    if verify:
        verify_pdf_url(url, 'Lancet')
    return url

def the_nature_ballet(pma, verify=True):
    '''  :param: pma (PubMedArticle object)
         :param: verify (bool) [default: True]
         :return: url (string)
         :raises: AccessDenied, NoPDFLink
    '''
    url = ''
    if pma.doi:
        try:
            starturl = the_doi_2step(pma.doi)
            url = starturl.replace('html', 'pdf').replace('abs', 'pdf').replace('full', 'pdf')
        except NoPDFLink:
            # alright, let's try the pii route.
            pass

    if url == '':
        if pma.pii:
            url = nature_format.format(a=pma, ja=nature_journals[standardize_journal_name(pma.journal)]['ja'])
        else:
            if pma.doi:
                raise NoPDFLink('MISSING: pii, TXERROR: dx.doi.org resolution failed for doi %s' % pma.doi)
            else:
                raise NoPDFLink('MISSING: pii, doi')

    if verify:
        verify_pdf_url(url, 'Nature')
    return url


PMC_PDF_URL = 'https://www.ncbi.nlm.nih.gov/pmc/articles/pmid/{a.pmid}/pdf'
EUROPEPMC_PDF_URL = 'http://europepmc.org/backend/ptpmcrender.fcgi?accid=PMC{a.pmc}&blobtype=pdf'
def the_pmc_twist(pma, verify=True, use_nih=False):
    '''Look up article in EuropePMC.org.  If not found there, fall back to NIH (if use_nih
    is True).

       Note: verification highly recommended. EuropePMC has most articles, but not everything.

         :param: pma (PubMedArticle object)
         :param: verify (bool) [default: True]
         :param: use_nih (bool) [default: False]
         :return: url
         :raises: NoPDFLink
    '''
    if pma.history.get('pmc-release', None):
        # marked "PAYWALL" here because FindIt will automatically retry such cached entries
        raise NoPDFLink('PAYWALL: pmc article in embargo until %s' % pma.history['pmc-release'].strftime('%Y-%m-%d'))

    url = EUROPEPMC_PDF_URL.format(a=pma)

    if not verify:
        return url

    try:
        verify_pdf_url(url, 'EuropePMC')
        return url

    except (NoPDFLink, AccessDenied):
        # Fallback to using NIH.gov if we're allowing it.        
        if use_nih:
            #   URL block might be discerned by grepping for this:
            #
            #   <div class="el-exception-reason">Bulk downloading of content by IP address [162.217...,</div>
            url = PMC_PDF_URL.format(a=pma)
            verify_pdf_url(url, 'NIH (EuropePMC fallback)')
            return url
    raise NoPDFLink('TXERROR: could not get PDF from EuropePMC.org and USE_NIH set to False')


def the_springer_shag(pma, verify=True):
    '''  :param: pma (PubMedArticle object)
         :param: verify (bool) [default: True]
         :return: url
         :raises: AccessDenied, NoPDFLink
    '''
    # start: http://link.springer.com/article/10.1007%2Fs13238-015-0153-5
    # return: http://link.springer.com/content/pdf/10.1007%2Fs13238-015-0153-5.pdf
    if pma.doi:
        baseurl = the_doi_2step(pma.doi)
    else:
        raise NoPDFLink('MISSING: doi (doi lookup failed)')
    url = baseurl.replace('article', 'content/pdf') + '.pdf'

    if verify:
        verify_pdf_url(url, 'Springer')
    return url

def the_karger_conga(pma, verify=True):
    '''  :param: pma (PubMedArticle object)
         :param: verify (bool) [default: True]
         :return: url
         :raises: AccessDenied, NoPDFLink
    '''
    # example: 23970213.  doi = 10.1159/000351538
    #       http://www.karger.com/Article/FullText/351538
    #       http://www.karger.com/Article/Pdf/351538

    if pma.doi: 
        if find_doi_in_string(pma.doi):
            kid = pma.doi.split('/')[1]
            if kid.isdigit():
                kid = str(int(kid))     # strips the zeroes that are padding the ID in the front.
        else:
            kid = pma.doi
            # sometimes the Karger ID was put in as the DOI (e.g. pmid 11509830)
        baseurl = 'http://www.karger.com/Article/FullText/%s' % kid
    else:
        raise NoPDFLink('MISSING: doi (doi lookup failed)')
    # if it directs to an "Abstract", we prolly can't get the PDF. Try anyway.
    url = baseurl.replace('FullText', 'Pdf').replace('Abstract', 'Pdf')

    if verify:
        verify_pdf_url(url, 'Karger')
    return url


def the_spandidos_lambada(pma, verify=True):
    '''  :param: pma (PubMedArticle object)
         :param: verify (bool) [default: True]
         :return: url
         :raises: AccessDenied, NoPDFLink
    '''
    jrnl = standardize_journal_name(pma.journal)
    baseurl = None
    try:
        pma = rectify_pma_for_vip_links(pma)  #raises NoPDFLink if missing data
        url = spandidos_format.format(ja=spandidos_journals[jrnl]['ja'], a=pma)
    except NoPDFLink:
        # let doi2step exceptions fall to calling function
        if pma.doi:
            baseurl = the_doi_2step(pma.doi)
            url = baseurl + '/download'
        else:
            raise NoPDFLink('MISSING: vip, doi - volume and/or issue missing from PubMedArticle; doi lookup failed.')

    if verify:
        verify_pdf_url(url, 'Spandidos')
    return url


def the_wolterskluwer_volta(pma, verify=True):
    '''  :param: pma (PubMedArticle object)
         :param: verify (bool) [default: True]
         :return: url
         :raises: AccessDenied, NoPDFLink
    '''
    doiurl = 'http://content.wkhealth.com/linkback/openurl?doi=%s'
    volissurl = 'http://content.wkhealth.com/linkback/openurl?issn={a.issn}&volume={a.volume}&issue={a.issue}&spage={a.first_page}'
    if pma.doi:
        baseurl = requests.get(doiurl % pma.doi).url
    elif pma.issn:
        pma = rectify_pma_for_vip_links(pma)  #raises NoPDFLink if missing data
        baseurl = requests.get(volissurl.format(a=pma)).url
        
    res = requests.get(baseurl)
    tree = etree.fromstring(res.content, HTMLParser())
    try:
        item = tree.cssselect('li.ej-box-01-body-li-article-tools-pdf')[0]
    except IndexError:
        raise NoPDFLink('DENIED: wolterskluwer did not provide PDF link for this article')
    link = item.getchildren()[0]
    url = link.get('href')
    if verify:
        verify_pdf_url(url)
    return url


def the_biochemsoc_saunter(pma, verify=True):
    '''  :param: pma (PubMedArticle object)
         :param: verify (bool) [default: True]
         :return: url
         :raises: AccessDenied, NoPDFLink
    '''
    jrnl = standardize_journal_name(pma.journal)
    pma = rectify_pma_for_vip_links(pma)  #raises NoPDFLink if missing data 
    host = biochemsoc_journals[jrnl]['host']
    url = biochemsoc_format.format(a=pma, host=host, ja=biochemsoc_journals[jrnl]['ja'])
    if verify:
        verify_pdf_url(url, 'biochemsoc')
    return url

def the_cell_pogo(pma, verify=True):
    '''  :param: pma (PubMedArticle object)
         :param: verify (bool) [default: True]
         :return: url
         :raises: AccessDenied, NoPDFLink
    '''
    jrnl = standardize_journal_name(pma.journal)
    if pma.pii:
        # the front door
        url = cell_format.format(a=pma, ja=cell_journals[jrnl]['ja'],
                                     pii=remove_chars(pma.pii, '-()'))
        if verify:
            verify_pdf_url(url, 'Cell')
        return url
    else:
        # let the SD function raise Exceptions
        return the_sciencedirect_disco(pma, verify)




from __future__ import absolute_import, unicode_literals

wiley_journals = [
        'APMIS',
        'Acad Emerg Med',
        'Acta Anaesthesiol Scand',
        'Acta Crystallogr C',
        'Acta Neurol Scand',
        'Acta Neurol Scand Supplc',
        'Acta Neuropsychiatr',
        'Acta Obstet Gynecol Scand',
        'Acta Obstet Gynecol Scand Suppl',
        'Acta Ophthalmol',
        'Acta Ophthalmol (Copenh)',
        'Acta Ophthalmol Scand',
        'Acta Ophthalmol Scand Suppl',
        'Acta Paediatr',
        'Acta Paediatr Jpn',
        'Acta Paediatr Scand',
        'Acta Paediatr Suppl',
        'Acta Pathol Jpn',
        'Acta Pathol Microbiol Scand B',
        'Acta Physiol (Oxf)',
        'Acta Physiol Scand',
        'Acta Psychiatr Scand',
        'Addict Biol',
        'Addiction',
        'Aging Cell',
        'Alcohol Clin Exp Res',
        'Aliment Pharmacol Ther',
        'Allergy',
        'Am J Addict',
        'Am J Anat',
        'Am J Hematol',
        'Am J Hum Biol',
        'Am J Ind Med',
        'Am J Med Genet',
        'Am J Med Genet A',
        'Am J Med Genet B Neuropsychiatr Genet',
        'Am J Med Genet C Semin Med Genet',
        'Am J Phys Anthropol',
        'Am J Primatol',
        'Am J Reprod Immunol',
        'Am J Transplant',
        'Anaesthesia',
        'Anat Histol Embryol',
        'Anat Rec',
        'Anat Rec (Hoboken)',
        'Andrologia',
        'Andrology',
        'Angew Chem Int Ed Engl',
        'Anim Genet',
        'Anim Sci J',
        'Ann Biol Clin (Paris)',
        'Ann Endocrinol (Paris)',
        'Ann Hum Genet',
        'Ann N Y Acad Sci',
        'Ann Neurol',
        'Ann Noninvasive Electrocardiol',
        'Antivir Ther (Lond)',
        'Arch Insect Biochem Physiol',
        'Arch Mal Coeur Vaiss',
        'Arch Pharm (Weinheim)',
        'Arthritis & rheumatology (Hoboken NJ)',
        'Arthritis Care Res (Hoboken)',
        'Arthritis Rheum',
        'Artif Organs',
        'Asia Pac J Clin Oncol',
        'Asia Pac Psychiatry',
        'Aust Dent J',
        'Aust Endod J',
        'Aust N Z J Obstet Gynaecol',
        'Aust Vet J',
        'Australas J Dermatol',
        'Autism Res',
        'BJOG',
        'BJU Int',
        'Basic Clin Pharmacol Toxicol',
        'Biochem Mol Biol Educ',
        'Biochem Mol Biol Int',
        'Bioelectromagnetics',
        'Bioessays',
        'Biofactors',
        'Biol Cell',
        'Biol Rev Camb Philos Soc',
        'Biom J',
        'Biomed Biochim Acta',
        'Biomed Chromatogr',
        'Bioorg Med Chem Lett',
        'Biopharm Drug Dispos',
        'Biopolymers',
        'Biotechnol Appl Biochem',
        'Biotechnol Bioeng',
        'Biotechnol J',
        'Biotechnol Prog',
        'Bipolar Disord',
        'Birth Defects Res A Clin Mol Teratol',
        'Birth Defects Res C Embryo Today',
        'Birth Defects Res Part A Clin Mol Teratol',
        'Br J Clin Pharmacol',
        'Br J Dermatol',
        'Br J Haematol',
        'Br J Surg',
        'Brain Pathol',
        'Breast J',
        'CNS Neurosci Ther',
        'Cancer',
        'Cancer Cytopathol',
        'Cancer Med',
        'Cancer Sci',
        'Cardiovasc Drug Rev',
        'Cardiovasc Ther',
        'Cas Lek Cesk',
        'Cathet Cardiovasc Diagn',
        'Catheter Cardiovasc Interv',
        'Cell Biochem Funct',
        'Cell Biol Int',
        'Cell Microbiol',
        'Cell Mol Biol (Noisy-le-grand)',
        'Cell Motil Cytoskeleton',
        'Cell Prolif',
        'Cell Tissue Kinet',
        'Chem Asian J',
        'Chem Biodivers',
        'Chem Biol Drug Des',
        'Chem Pharm Bull',
        'Chem Rec',
        'ChemMedChem',
        'ChemSusChem',
        'Chembiochem',
        'Chemistry',
        'Chemphyschem',
        'Child Dev',
        'Chirality',
        'Clin Cardiol',
        'Clin Endocrinol (Oxf)',
        'Clin Exp Allergy',
        'Clin Exp Dermatol',
        'Clin Exp Pharmacol Physiol',
        'Clin Exp Pharmacol Physiol Suppl',
        'Clin Experiment Ophthalmol',
        'Clin Genet',
        'Clin Lab Haematol',
        'Clin Oral Implants Res',
        'Clin Otolaryngol Allied Sci',
        'Clin Pharmacol Ther',
        'Clin Physiol',
        'Clin Physiol Funct Imaging',
        'Clin Respir J',
        'Clin Transl Sci',
        'Clin Transplant',
        'Cochrane Database Syst Rev',
        'Colorectal Dis',
        'Compr Physiol', 
        'Congenit Anom (Kyoto)',
        'Congenit Heart Dis',
        'Congest Heart Fail',
        'Contact Derm',
        'Contrast Media Mol Imaging',
        'Cytometry',
        'Cytometry A',
        'Cytometry B Clin Cytom',
        'Cytopathology',
        'Depress Anxiety',
        'Dermatol Ther',
        'Dev Dyn',
        'Dev Genet',
        'Dev Growth Differ',
        'Dev Med Child Neurol',
        'Dev Neurobiol',
        'Diabet Med',
        'Diabetes Metab Res Rev',
        'Diabetes Obes Metab',
        'Diagn Cytopathol',
        'Differentiation',
        'Dig Endosc',
        'Dis Esophagus',
        'Drug Test Anal',
        'EDTNA ERCA J',
        'Electrophoresis',
        'Emerg Med Australas',
        'Endod Dent Traumatol',
        'Environ Microbiol',
        'Environ Microbiol Rep',
        'Environ Mol Mutagen',
        'Environ Mutagen',
        'Environ Toxicol',
        'Environ Toxicol Chem',
        'Epilepsia',
        'Equine Vet J',
        'Equine Vet J Suppl',
        'Eur Eat Disord Rev',
        'Eur J Biochem',
        'Eur J Clin Invest',
        'Eur J Dermatol',
        'Eur J Gynaecol Oncol',
        'Eur J Haematol',
        'Eur J Haematol Suppl',
        'Eur J Heart Fail',
        'Eur J Immunogenet',
        'Eur J Immunol',
        'Eur J Neurol',
        'Eur J Neurosci',
        'Eur J Oral Sci',
        'Eur J Pain',
        'Evolution',
        'Exp Dermatol',
        'Exp Physiol',
        'FEBS J',
        'Farmaco',
        'Fundam Clin Pharmacol',
        'Genes Brain Behav',
        'Genes Cells',
        'Genes Chromosomes Cancer',
        'Genesis',
        'Genet Epidemiol',
        'Geobiology',
        'Geriatr Gerontol Int',
        'Gerodontology',
        'Glia',
        'HIV Med',
        'Haemophilia',
        'Head Neck',
        'Headache',
        'Helicobacter',
        'Hematol Oncol',
        'Hemodial Int',
        'Hepatology',
        'Hereditas',
        'Hippocampus',
        'Histopathology',
        'Hum Brain Mapp',
        'Hum Mutat',
        'Hum Psychopharmacol',
        'IUBMB Life',
        'Immunol Rev',
        'Influenza Other Respir Viruses',
        'Insect Mol Biol',
        'Insect Sci',
        'Int Arch Allergy Appl Immunol',
        'Int Dent J',
        'Int Endod J',
        'Int Forum Allergy Rhinol',
        'Int J Androl',
        'Int J Cancer',
        'Int J Clin Pract',
        'Int J Dermatol',
        'Int J Eat Disord',
        'Int J Geriatr Psychiatry',
        'Int J Gynecol Cancer',
        'Int J Immunogenet',
        'Int J Lab Hematol',
        'Int J Lang Commun Disord',
        'Int J Med Robot',
        'Int J Paediatr Dent',
        'Int J Pept Protein Res',
        'Int J Protein Res',
        'Int J Rheum Dis',
        'Int J Stroke',
        'Int J Urol',
        'Intern Med J',
        'J Am Assoc Nurse Pract',
        'J Am Geriatr Soc',
        'J Anat',
        'J Androl',
        'J Anim Breed Genet',
        'J Appl Bacteriol',
        'J Appl Microbiol',
        'J Appl Toxicol',
        'J Basic Microbiol',
        'J Biochem Mol Toxicol',
        'J Biolumin Chemilumin',
        'J Biomed Mater Res',
        'J Biomed Mater Res A',
        'J Biomed Mater Res Part B Appl Biomater',
        'J Bone Miner Res',
        'J Cardiometab Syndr',
        'J Cardiovasc Electrophysiol',
        'J Cell Biochem',
        'J Cell Biochem Suppl',
        'J Cell Mol Med',
        'J Cell Physiol',
        'J Chem Technol Biotechnol',
        'J Child Psychol Psychiatry',
        'J Clin Apher',
        'J Clin Hypertens (Greenwich)',
        'J Clin Lab Anal',
        'J Clin Periodontol',
        'J Clin Pharm Ther',
        'J Clin Pharmacol',
        'J Clin Psychol',
        'J Comp Neurol',
        'J Comput Chem',
        'J Cutan Pathol',
        'J Dermatol',
        'J Diabetes',
        'J Dig Dis',
        'J Dtsch Dermatol Ges',
        'J Eukaryot Microbiol',
        'J Eur Acad Dermatol Venereol',
        'J Evid Based Med',
        'J Exp Anal Behav',
        'J Exp Zool',
        'J Exp Zool A Ecol Genet Physiol',
        'J Exp Zool B Mol Dev Evol',
        'J Exp Zoolog Part A Comp Exp Biol',
        'J Fish Biol',
        'J Fish Dis',
        'J Food Sci',
        'J Forensic Sci',
        'J Formos Med Assoc',
        'J Gastroenterol Hepatol',
        'J Gene Med',
        'J Hepatobiliary Pancreat Sci',
        'J Hosp Med',
        'J Hum Nutr Diet',
        'J Immunogenet',
        'J Integr Plant Biol',
        'J Intellect Disabil Res',
        'J Intern Med',
        'J Lab Clin Med',
        'J Magn Reson Imaging',
        'J Mass Spectrom',
        'J Matern Fetal Med',
        'J Med Imaging Radiat Oncol',
        'J Med Primatol',
        'J Med Virol',
        'J Microsc',
        'J Mol Recognit',
        'J Morphol',
        'J Nanosci Nanotechnol',
        'J Neurobiol',
        'J Neurochem',
        'J Neuroendocrinol',
        'J Neurosci Res',
        'J New Drugs',
        'J Obstet Gynaecol Res',
        'J Oral Pathol Med',
        'J Oral Rehabil',
        'J Org Chem',
        'J Orthop Res',
        'J Paediatr Child Health',
        'J Pathol',
        'J Pept Res',
        'J Pept Sci',
        'J Periodont Res',
        'J Peripher Nerv Syst',
        'J Pharm Pharmacol',
        'J Pharm Sci',
        'J Physiol',
        'J Pineal Res',
        'J Prosthodont',
        'J Rural Health',
        'J Sch Health',
        'J Sci Food Agric',
        'J Sep Sci',
        'J Sex Med',
        'J Sleep Res',
        'J Small Anim Pract',
        'J Supramol Struct',
        'J Surg Oncol',
        'J Thromb Haemost',
        'J Tissue Eng Regen Med',
        'J Trauma Stress',
        'J Vet Emerg Crit Care (San Antonio)',
        'J Vet Intern Med',
        'J Vet Med A Physiol Pathol Clin Med',
        'J Vet Med B Infect Dis Vet Public Health',
        'J Vet Pharmacol Ther',
        'J Viral Hepat',
        'Jpn J Cancer Res',
        'Laryngoscope',
        'Lasers Surg Med',
        'Lett Appl Microbiol',
        'Liver',
        'Liver Int',
        'Liver Transpl',
        'Luminescence',
        'Macromol Biosci',
        'Macromol Rapid Commun',
        'Magn Reson Chem',
        'Magn Reson Med',
        'Med Educ',
        'Med Pediatr Oncol',
        'Med Res Rev',
        'Med Vet Entomol',
        'Ment Retard Dev Disabil Res Rev',
        'Microbiol Immunol',
        'Microcirculation',
        'Microsc Res Tech',
        'Mikrobiyol Bul',
        'Mol Carcinog',
        'Mol Ecol',
        'Mol Ecol Resour',
        'Mol Microbiol',
        'Mol Nutr Food Res',
        'Mol Oral Microbiol',
        'Mol Plant Pathol',
        'Mol Reprod Dev',
        'Mov Disord',
        'Mt Sinai J Med',
        'Muscle Nerve',
        'Muscle Nerve Suppl',
        'Mycoses',
        'NMR Biomed',
        'Nan Fang Yi Ke Da Xue Xue Bao',
        'Nat Toxins',
        'Nephrology (Carlton)',
        'Neurochem Res',
        'Neurogastroenterol Motil',
        'Neuropathol Appl Neurobiol',
        'Neuropathology',
        'Neurourol Urodyn',
        'New Phytol',
        'Nihon Shokakibyo Gakkai Zasshi',
        'Nippon Rinsho',
        'Novartis Found Symp',
        'Nutr Clin Care',
        'Obes Res',
        'Obes Rev',
        'Obesity (Silver Spring)',
        'Ophthalmic Physiol Opt',
        'Opt Express',
        'Oral Dis',
        'Oral Microbiol Immunol',
        'Orthod Craniofac Res',
        'Orthop Surg',
        'Otolaryngol Head Neck Surg',
        'Pacing Clin Electrophysiol',
        'Paediatr Anaesth',
        'Paediatr Perinat Epidemiol',
        'Pain Med',
        'Pathol Int',
        'Pediatr Allergy Immunol',
        'Pediatr Blood Cancer',
        'Pediatr Dermatol',
        'Pediatr Diabetes',
        'Pediatr Int',
        'Pediatr Obes',
        'Pediatr Transplant',
        'Pest Manag Sci',
        'Pharmacoepidemiol Drug Saf',
        'Pharmacol Toxicol',
        'Pharmacotherapy',
        'Pharmazie',
        'Photochem Photobiol',
        'Photodermatol Photoimmunol Photomed',
        'Physiol Plant',
        'Phytochem Anal',
        'Phytother Res',
        'Pigment Cell Melanoma Res',
        'Pigment Cell Res',
        'Plant Biol (Stuttg)',
        'Plant Biotechnol J',
        'Plant Cell Environ',
        'Plant J',
        'Prenat Diagn',
        'Proc Assoc Am Physicians',
        'Prostate',
        'Protein Sci',
        'Proteins',
        'Proteomics',
        'Proteomics Clin Appl',
        'Psychiatry Clin Neurosci',
        'Psychogeriatrics',
        'Psychophysiology',
        'Public Health Nurs',
        'Rapid Commun Mass Spectrom',
        'Recenti Prog Med',
        'Reprod Domest Anim',
        'Respirology',
        'Rev Med Virol',
        'Scand J Haematol',
        'Scand J Immunol',
        'Scand J Med Sci Sports',
        'Skin Res Technol',
        'Small',
        'Stat Med',
        'Stem Cells',
        'Synapse',
        'Teratog Carcinog Mutagen',
        'Teratology',
        'Ther Apher Dial',
        'Tissue Antigens',
        'Traffic',
        'Transbound Emerg Dis',
        'Transfus Med',
        'Transfusion',
        'Transpl Infect Dis',
        'Transpl Int',
        'Trop Med Int Health',
        'Ugeskr Laeg',
        'Ultrasound Obstet Gynecol',
        'Verh Dtsch Ges Pathol',
        'Vet Anaesth Analg',
        'Vet Clin Pathol',
        'Vet Dermatol',
        'Vet Ophthalmol',
        'Vet Surg',
        'Vox Sang',
        'Wiad Lek',
        'Wiley Interdiscip Rev RNA',
        'Wound Repair Regen',
        'Xenotransplantation',
        'Yakugaku Zasshi',
        'Yeast',
        'Yi Chuan Xue Bao',
        'Z Allg Mikrobiol',
        'Z Tierpsychol',
        'Zentralbl Veterinarmed A',
        'Zentralblatt Veterinarmedizin Reihe B',
        'Zh Nevrol Psikhiatr Im S S Korsakova',
        'Zhongguo Dang Dai Er Ke Za Zhi',
        'Zhongguo Shi Yan Xue Ye Xue Za Zhi',
        'Zhongguo Yi Xue Ke Xue Yuan Xue Bao',
        'Zhonghua Bing Li Xue Za Zhi',
        'Zhonghua Er Ke Za Zhi',
        'Zhonghua Kou Qiang Yi Xue Za Zhi',
        'Zhonghua Yan Ke Za Zhi',
        'Zoonoses Public Health',
]



from __future__ import print_function, absolute_import, unicode_literals

import logging

import requests

from .eutils_common import SQLiteCache, get_cache_path
from .base import Borg
from .config import DEFAULT_CACHE_DIR
from .exceptions import BadDOI, DxDOIError
from .text_mining import find_doi_in_string

DX_DOI_URL = 'http://dx.doi.org/%s'
CACHE_FILENAME = 'dx_doi-cache.db'

DX_DOI_CACHE = None

def _get_dx_doi_cache(cachedir=DEFAULT_CACHE_DIR):
    global DX_DOI_CACHE
    if not DX_DOI_CACHE:
        _cache_path = get_cache_path(cachedir, CACHE_FILENAME)
        DX_DOI_CACHE = SQLiteCache(_cache_path)
    return DX_DOI_CACHE


class DxDOI(Borg):
    """ Looks up DOIs in dx.doi.org and caches results in an SQLite
    cache. This is a Borg singleton object.

    Methods:

        resolve (doi, *args): uses supplied doi to get link to publisher.

        check_doi (doi, *args): returns doi if supplied DOI is good,
                                raises BadDOI if not good.
    """

    _log = logging.getLogger('metapub.DxDOI')

    def __init__(self, **kwargs):
        if kwargs.get('debug', False):
            self._log.setLevel(logging.DEBUG)
        else:
            self._log.setLevel(logging.INFO)

        cachedir = kwargs.get('cachedir', DEFAULT_CACHE_DIR)
        self._cache = _get_dx_doi_cache(cachedir)

    def check_doi(self, doi, whitespace=False):
        """ Checks validity of supplied doi.

        If whitespace is True (default False), allows supplied doi to
        contain whitespace.

        :param doi: (str)
        :param whitespace: (bool)
        :return: doi (str) -- verified DOI)
        :raise BadDOI if supplied DOI fails regular expression check.
        """

        result_doi = find_doi_in_string(doi, whitespace=False)
        if result_doi is None:
            raise BadDOI('Supplied DOI "%s" fails doi check' % doi)
        return doi

    def _query_api(self, doi):
        response = requests.get(DX_DOI_URL % doi)
        if response.status_code in [200, 401, 301, 302, 307, 308, 416]:
            return response.url
        else:
            raise DxDOIError('dx.doi.org lookup failed for doi "%s" (HTTP %i returned)' %
                            (doi, response.status_code))

    def resolve(self, doi, check_doi=True, whitespace=False, skip_cache=False):
        """ Takes a doi (string), returns a url to article page on journal website.

        if check_doi is True (default True), checks DOI before
        submitting query to dx.doi.org.

        if whitespace is True (default False), allows prospective
        dois to contain whitespace when checked.

        if skip_cache is True (default False), doesn't check cache for
        pre-existing results (loads from remote dx.doi.org).

        :param doi: (str)
        :param check_doi: (bool)
        :param whitespace: (bool)
        :param skip_cache: (bool)
        :return: url (str)
        :raises BadDOI: if supplied DOI failed regular expression check
        :raises DxDOIError: if not-ok HTTP status code while loading url
        :raises ConnectionError: if problem making dx.doi.org connection
        """
        if doi is None or doi.strip()=='':
            raise BadDOI('DOI cannot be None or empty string')

        if check_doi:
            doi = self.check_doi(doi, whitespace=whitespace)
            
        url = None
        if not skip_cache:
            url = self._query_cache(doi)

        if url == None:
            url = self._query_api(doi)

            if self._cache:
                cache_key = self._make_cache_key(doi)
                self._cache[cache_key] = url
                self._log.info('cached results for key {cache_key} ({doi}) '.format(
                        cache_key=cache_key, doi=doi))
        return url

    def _make_cache_key(self, inp):
        return inp.strip()

    def _query_cache(self, key):
        """ Return results for a cache lookup, if found.

        :param key: (str)
        :return: val (str) or None
        """
        if self._cache:
            cache_key = self._make_cache_key(key)
            try:
                val = self._cache[cache_key]
                self._log.debug('cache hit for key {cache_key} ({key}) '.format(
                    cache_key=cache_key, key=key))
                return val
            except KeyError:
                self._log.debug('cache miss for key {cache_key} ({key}) '.format(
                        cache_key=cache_key, key=key))
                return None
        else:
            self._log.debug('cache disabled (self._cache is None)')
            return None



from __future__ import absolute_import, unicode_literals

simple_formats_pii = {
    'Am Heart J': 'http://www.ahjonline.com/article/{a.pii}/pdf',
    'Am J Cardiol': 'http://www.ajconline.org/article/{a.pii}/pdf',
    'Am J Ophthalmol': 'http://www.ajo.com/article/{a.pii}/pdf',
    'Am J Med': 'http://www.amjmed.com/article/{a.pii}/pdf',
    'Atherosclerosis': 'http://www.atherosclerosis-journal.com/article/{a.pii}/pdf',
    'Arch Med Res': 'http://www.arcmedres.com/article/{a.pii}/pdf',
    'Biol Psychiatry': 'http://www.biologicalpsychiatryjournal.com/article/{a.pii}/pdf',
    'Bone': 'http://www.thebonejournal.com/article/{a.pii}/pdf',
    'Brain Dev': 'http://www.brainanddevelopment.com/article/{a.pii}/pdf',
    'Cancer Cell Int': 'http://www.cancerci.com/content/pdf/{a.pii}.pdf',
    'Cancer Genet Cytogenet': 'http://www.cancergeneticsjournal.org/article/{a.pii}/pdf',
    'Cancer Lett': 'http://www.cancerletters.info/article/{a.pii}/pdf',
    'Clin Neurol Neurosurg': 'http://www.clineu-journal.com/article/{a.pii}/pdf',
    'Diabetes Res Clin Pract': 'http://www.diabetesresearchclinicalpractice.com/article/{a.pii}/pdf',
    'Epilepsy Res': 'http://www.epires-journal.com/article/{a.pii}/pdf',
    'Eur J Paediatr Neurol': 'http://www.ejpn-journal.com/article/{a.pii}/pdf',
    'Exp Hematol': 'http://www.exphem.org/article/{a.pii}/pdf',
    'Fertil Steril': 'http://www.fertstert.org/article/{a.pii}/pdf',
    'Gastroenterology': 'http://www.gastrojournal.org/article/{a.pii}/pdf',
    'Gynecol Oncol': 'http://www.gynecologiconcology-online.net/article/{a.pii}/pdf',
    'Heart Rhythm': 'http://www.heartrhythmjournal.com/article/{a.pii}/pdf',
    'Int J Pediatr Otorhinolaryngol': 'http://www.ijporlonline.com/article/{a.pii}/pdf',
    'Int J Cardiol': 'http://www.internationaljournalofcardiology.com/article/{a.pii}/pdf',
    'J Dermatol': 'http://www.jdsjournal.com/article/{a.pii}/pdf',
    'J Mol Cell Cardiol': 'http://www.jmmc-online.com/article/{a.pii}/pdf',
    'J Mol Diagn': 'http://jmd.amjpathol.org/article/{a.pii}/pdf',
    'J Neurol Sci': 'http://www.jns-journal.com/article/{a.pii}/pdf',
    'J Pediatr': 'http://www.jpeds.com/article/{a.pii}/pdf',
    'J Pediatr Urol': 'http://www.jpurol.com/article/{a.pii}/pdf',
    'Ophthalmology': 'http://www.aaojournal.org/article/{a.pii}/pdf',
    'Orv Hetil': 'http://www.akademiai.com/content/{a.pii}/fulltext.pdf',
    'Med Hypotheses': 'http://www.medical-hypotheses.com/article/{a.pii}/pdf',
    'Metabolism': 'http://www.metabolismjournal.com/article/{a.pii}/pdf',
    'Metab Clin Exp': 'http://www.metabolismjournal.com/article/{a.pii}/pdf',
    'Mol Genet Metab': 'http://www.mgmjournal.com/article/{a.pii}/pdf',
    'Neurobiol Aging': 'http://www.neurobiologyofaging.org/article/{a.pii}/pdf',
    'Neuromuscul Disord': 'http://www.nmd-journal.com/article/{a.pii}/pdf',
    'Parkinsonism Relat Disord': 'http://www.prd-journal.com/article/{a.pii}/pdf',
    'Pediatr Neurol': 'http://www.pedneur.com/article/{a.pii}/pdf',
    'Placenta': 'http://www.placentajournal.org/article/{a.pii}/pdf',
    'Surg Neurol': 'http://www.worldneurosurgery.org/article/{a.pii}/pdf',
    'Thromb Res': 'http://www.thrombosisresearch.com/article/{a.pii}/pdf',
}



from __future__ import absolute_import, unicode_literals

# JSTAGE: mostly free (yay)
# Examples:
# J Biochem: https://www.jstage.jst.go.jp/article/biochemistry1922/125/4/125_4_803/_pdf
# Drug Metab Pharmacokinet:
# https://www.jstage.jst.go.jp/article/dmpk/20/2/20_2_144/_article -->
# https://www.jstage.jst.go.jp/article/dmpk/20/2/20_2_144/_pdf
jstage_journals = [
        'Agric Biol Chem',
        'Anal Sci',
        'Ann Thorac Cardiovasc Surg',
        'Biol Pharm Bull',
        'Biomed Res',
        'Biosci Trends',
        'Cell Struct Funct',
        'Chem Pharm Bull',
        'Circ J',
        'Crit Rev Immunol',
        'Dent Mater J',
        'Drug Metab Pharmacokinet',
        'Endocr J',
        'Exp Anim',
        'Fukushima J Med Sci',
        'Genes Genet Syst',
        'Int Heart J',
        'Intern Med',
        'J Atheroscler Thromb',
        'J Biochem',
        'J Clin Exp Hematop',
        'J Epidemiol',
        'J Gen Appl Microbiol',
        'J Med Invest',
        'J Nippon Med Sch',
        'J Nutr Sci Vitaminol',
        'J Occup Health',
        'J Oleo Sci',
        'J Pharmacobio-dyn',# bpb1978
        'J Reprod Dev',     # jrd
        'J Toxicol Sci',    # jts
        'J Vet Med Sci',    # jvms
        'Jpn Circ J',       # jcj
        'Jpn Heart J',      # jhj
        'Jpn J Infect Dis',
        'Jpn J Med',
        'Jpn J Med Sci Biol',
        'Jpn J Pharmacol',
        'Jpn J Physiol',
        'Kansenshogaku Zasshi',
        'Keio J Med',
        'Kurume Med J',
        'Microbes Environ',
        'Neurol Med Chir (Tokyo)',
        'Nihon Hansenbyo Gakkai Zasshi',
        'Nihon Ishinkin Gakkai Zasshi',
        'Nihon Naibunpi Gakkai Zasshi',
        'Nihon Rinsho Meneki Gakkai Kaishi',
        'Nihon Ronen Igakkai Zasshi',
        'Nihon Shishubyo Gakkai Kaishi',
        'Nippon Ika Daigaku Zasshi',
        'Nippon Jibiinkoka Gakkai Kaiho',
        'Nippon Juigaku Zasshi',
        'Nippon Naika Gakkai Zasshi',
        'Nippon Saikingaku Zasshi',
        'Nippon Yakurigaku Zasshi',
        'Rinsho Ketsueki',
        'Rinsho Shinkeigaku',
        'Sangyo Igaku',
        'Shokuhin Eiseigaku Zasshi',
        'Tohoku J Exp Med',
        'Yakugaku Zasshi',
]



# -*- encoding: utf-8 -*-

from __future__ import unicode_literals

from metapub.utils import asciify

# leet ipsum should not change at all (our "control").
leet_ipsum = "1F w1t t3xt r35ul7. |3tz0rz, |1nk v13w 15 aLL. T3H be b|00 kl1x r3zUltz, y0 7he 534r(h (83t4). D0 54y, 0f73n 5peNDInG iTz, != +HE p4r7I 4lvvAyz. FOr w1ll v3r510n t0. (l1(k z3aRcH w1+hOUT 1T 4nd, 47 why p1>< HELp 34513r. Up +h3 m1t3 |23p|4c3d."
leet_result = b"1F w1t t3xt r35ul7. |3tz0rz, |1nk v13w 15 aLL. T3H be b|00 kl1x r3zUltz, y0 7he 534r(h (83t4). D0 54y, 0f73n 5peNDInG iTz, != +HE p4r7I 4lvvAyz. FOr w1ll v3r510n t0. (l1(k z3aRcH w1+hOUT 1T 4nd, 47 why p1>< HELp 34513r. Up +h3 m1t3 |23p|4c3d."

# the following two samples have international alphabet characters that should get stripped out.
german_ipsum = "Rëm un Hunn duurch, dir Ierd ston dé. As déi alle séngt Friemd, fu sin hale bléit. Un fond schnéiwäiss aus. Ke wait hinnen beschéngt mir. Hun un hire zielen d'Stroos, ze get schlon klinzecht, Mecht Hämmel wee si. Léift schéinste gin de, dénen d'Mier de blo."
german_result = b"Rm un Hunn duurch, dir Ierd ston d. As di alle sngt Friemd, fu sin hale blit. Un fond schniwiss aus. Ke wait hinnen beschngt mir. Hun un hire zielen d'Stroos, ze get schlon klinzecht, Mecht Hmmel wee si. Lift schinste gin de, dnen d'Mier de blo."

volapuk_ipsum = "Ot evisitobs-li ziläk ifi. Löf nästön ospaloms sevön me."
volapuk_result = b"Ot evisitobs-li zilk ifi. Lf nstn ospaloms sevn me."

# non-ascii glyphs should get stripped out.
glyph_samples = {"¡ <-- upsidedown exclamation mark.": b" <-- upsidedown exclamation mark.",
                 "© is a copyright sign": b" is a copyright sign",
                 "® is a registered trademark": b" is a registered trademark",
                 "¾ <-- vulgar three-quarters": b" <-- vulgar three-quarters",
                 "the alpha (α) and the omega (ω)": b"the alpha () and the omega ()",
                 }


def test_glyphs():
    for glyph in list(glyph_samples.keys()):
        result = asciify(glyph)
        assert type(result) == bytes
        assert result == glyph_samples[glyph]


def test_german_ipsum():
    result = asciify(german_ipsum)
    assert type(result) == bytes
    assert result == german_result


def test_leet_ipsum():
    result = asciify(leet_ipsum)
    assert type(result) == bytes
    assert result == leet_result


def test_volapuk_ipsum():
    result = asciify(volapuk_ipsum)
    assert type(result) == bytes
    assert result == volapuk_result


if __name__=='__main__':
    test_german_ipsum()
    test_leet_ipsum()
    test_volapuk_ipsum()



from __future__ import absolute_import, unicode_literals

import re

import requests

try:
    from urlparse import urlparse
except ImportError:
    # assume python3
    from urllib.parse import urlparse

from .utils import remove_html_markup
    

# examples of real DOIs: 
#         10.1002/(SICI)1098-1004(1999)14:1<91::AID-HUMU21>3.0.CO;2-B
#         10.1007/s12020-014-0368-x
#         10.1101/gad.14.3.278
#         10.1016/S0898-1221(00)00204-2

# example of DOI returned by the regular expression that is incorrect:
#         10.1002/ajmg.b.30585).

re_doi = re.compile(r'(10[.][0-9]{2,}(?:[.][0-9]+)*/(?:(?!["&\'])\S)+)')
re_doi_ws = re.compile(r'(10[.][0-9]{2,}(?:[.][0-9]+)*\s+/\s+(?:(?!["&\'])\S)+)')

re_pmid = re.compile('\d+')
re_numbers = re_pmid    # for now, until there's a better idea about parsing PMIDs...


def pick_pmid(text):
    """ return longest numerical string from text (string) as the pmid.
        if text is empty or there are no pmids, return None.

    :param text: (str)
    :return: pmid (str) or None
    """
    pmids = re_pmid.findall(text)
    if pmids:
        longest = ''
        for num in pmids:
            if len(num) > len(longest):
                longest = num
        return longest
    else:
        return None


def _doi_pass_2(doi):
    if doi.endswith('.') or doi.endswith(','):
        return _doi_pass_2(doi[:-1])
    elif doi.endswith(')'):
        if doi.count('(') == doi.count(')'):
            return doi
        else:
            return _doi_pass_2(doi[:-1])
    else:
        return doi.replace(' ', '')


def findall_dois_in_text(inp, whitespace=False):
    """ Returns all seen DOIs in submitted text.

        if `whitespace` arg set to True, look for DOIs like the following:
             10.1002 / pd.354
        ...but return with whitespace stripped:
             10.1002/pd.354

    :param inp: (str)
    :param whitespace: (bool)
    :return: list of DOIs found in inp
    """
    if whitespace:
        dois = re_doi_ws.findall(inp)
    else:
        dois = re_doi.findall(inp)
    return [_doi_pass_2(doi) for doi in dois]


def find_doi_in_string(inp, whitespace=False):
    """ Returns the first seen DOI in the input string.

    :param inp: (str)
    :param whitespace: (bool)
    :return: string containing first found DOI, or None
    """
    try:
        doi = findall_dois_in_text(inp, whitespace)[0]
    except IndexError:
        return None
    return _doi_pass_2(doi)


def scrape_doi_from_article_page(url):
    """ Takes an article link (url), loads its page, and searches its content for DOIs, returning
    the first one it finds.

    The first DOI found on the page being the correct one for the article at hand seems to be a
    reasonable and workable assumption in general.

    :param url: (str)
    :return: doi or None
    """
    response = requests.get(url)
    if response.ok:
        dois = findall_dois_in_text(response.text)
        if dois:
            return remove_html_markup(dois[0])
    return None


# NOT TESTED and probably not working #####
def get_pmc_fulltext_filename_for_PubMedArticle(pma):
    fmt = '{journal}/{journal}_{year}_{month}_{day}_{voliss}_{pages}'
    return fmt.format(**pma.to_dict())



from __future__ import absolute_import, unicode_literals, print_function

from pkgutil import extend_path
__path__ = extend_path(__path__, __name__)

from .pubmedarticle import PubMedArticle
from .pubmedauthor import PubMedAuthor
from .pubmedfetcher import PubMedFetcher
from .medgenfetcher import MedGenFetcher
from .medgenconcept import MedGenConcept
from .clinvarfetcher import ClinVarFetcher
from .clinvarvariant import ClinVarVariant
from .crossref import CrossRef
from .findit import FindIt
from .dx_doi import DxDOI
from .urlreverse import UrlReverse

__version__ = '0.4.3.5'




from __future__ import print_function, absolute_import, unicode_literals

__doc__='mildly-experimental mashups of various services to get needed IDs.'

from .pubmedfetcher import PubMedFetcher
from .crossref import CrossRef
from .exceptions import *

crossref = None   #CrossRef()
pm_fetch = None   #PubMedFetcher()

def _start_engines():
    global crossref
    global pm_fetch
    if crossref and pm_fetch:
        return
    else:
        crossref = CrossRef()
        pm_fetch = PubMedFetcher()


def _protected_crossref_query(**kwargs):
    pma = kwargs.get('pma', None)
    pmid = kwargs.get('pmid', None)
    use_best_guess = kwargs.get('use_best_guess', False)
    min_score = kwargs.get('min_score', 2.0)
    if pmid and not pma:
        pma = pm_fetch.article_by_pmid(pmid)
    try:
        results = crossref.query_from_PubMedArticle(pma)
    except CrossRefConnectionError:
        return None
    return crossref.get_top_result(results, crossref.last_params, use_best_guess, min_score=min_score)


def interpret_pmids_for_citation_results(pmids):
    if len(pmids) == 1:
        if pmids[0] == 'NOT_FOUND':
            return None
        elif pmids[0].startswith('AMBIGUOUS'):
            return 'AMBIGUOUS'
        return str(pmids[0])
    elif len(pmids) == 0:
        return None
    else:
        return 'AMBIGUOUS'


def PubMedArticle2doi(pma, use_best_guess=False, min_score=2.0):
    '''starting with a PubMedArticle object, use CrossRef to find a DOI for given article.

    Args:
        pma (PubMedArticle)
        use_best_guess (bool): default=False
        min_score (float): default=2.0

    Returns:
        doi (str) or None
    '''
    _start_engines()
    result = _protected_crossref_query(pma=pma, use_best_guess=use_best_guess, min_score=min_score)
    if result:
        return result['doi']
    else:
        return None


def PubMedArticle2doi_with_score(pma, use_best_guess=False, min_score=2.0):
    '''Starting with a PubMedArticle object, use CrossRef to find a DOI for given article.

    Returns a tuple containing the DOI and the score CrossRef returned for the
    lookup.  If there was no good result above the min_score threshold, the tuple 
    will contain (None, 0.0).

    Args:
        pma (PubMedArticle)
        use_best_guess (bool): default=False
        min_score (float): default=2.0
    
    Returns:
        tuple (doi string, score) or (None, 0.0)
    '''

    _start_engines()
    result = _protected_crossref_query(pma=pma, use_best_guess=use_best_guess, min_score=min_score)
    if result:
        return (result['doi'], result['score'])
    else:
        return (None, 0.0)


def pmid2doi(pmid, use_best_guess=False, min_score=2.0):
    '''starting with a pubmed ID, lookup article in pubmed. If DOI found in PubMedArticle object,
        return it.  Otherwise, use CrossRef to find the DOI for given article.

    Args:
        pmid (str or int)
        use_best_guess (bool): default=False
        min_score (float): minimum score to accept from CrossRef for given doi. default=2.0

    Returns:
        doi (str) or None

    Raises:
        InvalidPMID (if pmid is invalid)
    '''
    # let MetaPubError pass back to the caller if pmid is not for realz..
    _start_engines()
    pma = pm_fetch.article_by_pmid(pmid)
    if pma.doi:
        return pma.doi
    return PubMedArticle2doi(pma, use_best_guess, min_score=min_score)


def pmid2doi_with_score(pmid, use_best_guess=False, min_score=2.0):
    '''Starting with a pubmed ID, lookup article in pubmed. 
    
    If DOI found in PubMedArticle object, that doi and 10.0 for the doi_score, 
    i.e. the tuple (doi, 10.0).

    Otherwise, use CrossRef to find the DOI for given article and return (doi, crossref_doi_score).

    Args:
        pmid (str or int)
        use_best_guess (bool): default=False
        min_score (float): minimum score to accept from CrossRef for given doi. default=2.0

    Returns:
        tuple (doi, score) or (None, 0.0)

    Raises:
        InvalidPMID (if pmid is invalid)
    '''
    _start_engines()
    pma = pm_fetch.article_by_pmid(pmid)
    if pma.doi:
        return (pma.doi, 10.0)
    return PubMedArticle2doi_with_score(pma, use_best_guess, min_score=min_score)


def doi2pmid(doi, use_best_guess=False, min_score=2.0, debug=False):
    '''uses CrossRef and PubMed eutils to lookup a PMID given a known doi.

    Warning: NO validation of input DOI performed here. Use
             metapub.text_mining.find_doi_in_string beforehand if needed.

    If a PMID can be found, return it. Otherwise return None.

    In very rare cases, use of the CrossRef->pubmed citation method used
    here may result in more than one pubmed ID. In this case, this function
    will return instead the word 'AMBIGUOUS'.

    Args:
        pmid (str or int)
        use_best_guess (bool): default=False
        min_score (float): minimum score to accept from CrossRef for given doi. default=2.0

    Returns:
        pmid (str) if found; 'AMBIGUOUS' if citation count > 1; None if no results.
    '''
    # for PMA, skip the validation; some pubmed XML has weird partial strings for DOI.
    # We should allow people to search using these oddball strings.
    _start_engines()
    doi = doi.strip()
    try:
        pma = pm_fetch.article_by_doi(doi)
        if debug:
            print('Found PubMedArticle via eutils fetch')
        return pma.pmid
    except:
        pass

    # Try doing a DOI lookup right in an advanced query string. Sometimes works and has
    # benefit of being a cached query so it is quick to do again, should we need.
    pmids = pm_fetch.pmids_for_query(doi)
    if len(pmids) == 1:
        # we need to cross-check; pubmed sometimes screws us over by giving us an article
        # with a SIMILAR doi. *facepalm*
        pma = pm_fetch.article_by_pmid(pmids[0])
        if pma.doi == doi:
            if debug:
                print('Found PMID via PubMed advanced query for DOI')
            return pma.pmid
        if debug:
            print('PubMed advanced query gave us a wonky result:')
            print('     Search: %s' % doi)
            print('     Return: %s' % pma.doi)

    # Look up the DOI in CrossRef, then feed results to pubmed citation query tool.
    try:
        results = crossref.query(doi)
    except CrossRefConnectionError:
        return None

    if results:
        top_result = crossref.get_top_result(results, crossref.last_params, use_best_guess, min_score=min_score)
        pmids = pm_fetch.pmids_for_citation(debug=debug, **top_result['slugs'])
        if debug:
            print('Submitted slugs: %r' % top_result['slugs'])
            #print('CrossRef results: %r' % top_result)
            print('PMIDs: %r' % pmids)
        return interpret_pmids_for_citation_results(pmids)
    else:
        return None




import unittest, os

from metapub import MedGenFetcher

hugos = ['ACVRL1', 'FOXP3', 'ATM']

TEST_CACHEDIR = 'tests/testcachedir'


class TestMedGenFetcher(unittest.TestCase):

    def setUp(self):
        self.fetch = MedGenFetcher()

    def tearDown(self):
        pass

    def test_fetch_concepts_for_known_gene(self):
        hugo = 'ACVRL1'
        result = self.fetch.uids_by_term(hugo+'[gene]')
        assert result is not None
        assert result[0] == '324960'
    
    def test_fetch_concepts_for_incorrect_term(self):
        term = 'AVCRL'
        result = self.fetch.uids_by_term(term+'[gene]')
        assert result == []

    def test_configurable_cachedir(self):
        """ Test that `cachedir` keyword argument is fully supported in modes:

        cachedir == 'default'   <-- assumed working since other tests use this.
        cachedir is None
        cachedir is 'some/path'
        cachedir is '~/path'
        """

        cachedir = TEST_CACHEDIR
        # start with cachedir==None; test that no cachedir is created.
        fetch = MedGenFetcher(cachedir=None)
        assert not os.path.exists(cachedir)

        fetch = MedGenFetcher(cachedir=cachedir)
        assert os.path.exists(cachedir)

        os.unlink(fetch._cache_path)
        os.rmdir(cachedir)

        fetch = MedGenFetcher(cachedir='~/testcachedir')
        assert os.path.exists(os.path.expanduser('~/testcachedir'))

        os.unlink(fetch._cache_path)
        os.rmdir(os.path.expanduser('~/testcachedir'))



import unittest

from metapub import PubMedFetcher
fetch = PubMedFetcher()

# fixtures
NOT_FOUND_INVALID_JOURNAL_params = {'jtitle': 'Computers',
                                    'year': 2000,
                                    'volume': 40,
                                    'spage': 885
                                    }

many_authors_params = {'jtitle': 'American Journal of Medical Genetics',
                       'year': 1996,
                       'volume': 61,
                       'spage': 10,
                       'authors': 'Katherine M. Hegmann; Aimee S. Spikes; Avi Orr-Urtreger; Lisa G. Shaffer'
                       }

no_authors_params = {'jtitle': 'Journal of Neural Transmission',
                     'year': 2014,
                     'volume': 121,
                     'first_page': 1077,
                     }

# PNAS|2008|||An|metapub|AMBIGUOUS (5 citations)
AMBIGUOUS_params = {'jtitle': 'PNAS', 'year': 2008, 'aulast': 'An'}


class TestPubmedCitationMatch(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_citation_match_without_authors(self):
        result = fetch.pmids_for_citation(**no_authors_params)
        assert result[0] == u'25023161'

    def test_citation_match_with_many_authors(self):
        result = fetch.pmids_for_citation(**many_authors_params)
        assert result[0] == u'8741910'

    # def test_citation_match_not_found(self):
    #    result = fetch.pmids_for_citation(**NOT_FOUND_params)
    #    assert result[0] == u'NOT_FOUND'

    def test_citation_match_not_found_invalid_journal(self):
        result = fetch.pmids_for_citation(**NOT_FOUND_INVALID_JOURNAL_params)
        assert result[0] == u'NOT_FOUND;INVALID_JOURNAL'

    def test_citation_match_ambiguous(self):
        result = fetch.pmids_for_citation(**AMBIGUOUS_params)
        assert result[0] == u'AMBIGUOUS (5 citations)'



from __future__ import absolute_import, unicode_literals

import unittest
import os

from metapub.base import parse_elink_response

from .common import CWD

datadir = os.path.join(CWD, 'data')


class TestConversions(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_medgen_pubmed_elink_response_0_ids(self):
        fixture = open(os.path.join(datadir, 'medgen_elink_pubmed_0_entries.xml')).read()
        ids = parse_elink_response(fixture)
        assert len(ids) == 0




from __future__ import absolute_import, unicode_literals

import os, logging

# this import is used by other modules (but not in this one):
from eutils.sqlitecache import SQLiteCache

from .config import DEFAULT_EMAIL, PKGNAME
from .exceptions import MetaPubError

EUTILS_DEFAULT_CACHEDIR = os.path.expanduser('~/.cache')

# == suppress the chatter of eutils and requests == #
logging.getLogger('eutils').setLevel(logging.WARNING)
logging.getLogger('requests').setLevel(logging.ERROR)
# ==


def get_eutils_client(cache_path, email=DEFAULT_EMAIL):
    """
    :param cache_path: valid filesystem path to SQLite cache file
    :param email: (optional) email address to submit with cache queries
    :return: eutils QueryService client object
    """
    import eutils.client as ec
    if cache_path is None:
        return ec.QueryService(tool=PKGNAME, email=email)   #, cache_path=None)
    return ec.QueryService(tool=PKGNAME, email=email)   #, cache_path=cache_path)


def get_cache_path(cachedir, filename='metapub-cache.db'):
    """ checks if cachedir exists; if not, tries to create it;
    raises MetaPubError if it can't be created.

        if cachedir is None, returns None.
        if cachedir is 'default', returns the eutils default of '~/.cache/<filename>'

        Supports expansion of user directory shortcut '~' to full path.

    :param cachedir: directory to store
    :param filename: name of cache file
    :return: path to SQLite DB file
    :raises MetaPubError
    """
    if cachedir is None:
        return None

    elif cachedir == 'default':
        cachedir = EUTILS_DEFAULT_CACHEDIR

    elif cachedir.find('~') > -1:
        cachedir = os.path.expanduser(cachedir)

    if _require_dir(cachedir):
        return os.path.join(cachedir, filename)
    else:
        raise MetaPubError('Could not create cache directory location %s' % cachedir)


def _require_dir(targetdir):
    if os.path.exists(targetdir):
        return True

    try:
        os.makedirs(targetdir)
        return True
    except OSError:
        return False



from __future__ import absolute_import, unicode_literals

# nature journals -- see http://www.nature.com/siteindex/index.html
nature_format = 'http://www.nature.com/{ja}/journal/v{a.volume}/n{a.issue}/pdf/{a.pii}.pdf'
nature_journals = {
    'Am J Gastroenterol': {'ja': 'ajg'},
    'Eur J Clin Nutr': {'ja': 'ejcn'},
    'Eur J Hum Genet': {'ja': 'ejhg'},
    'Eye (Lond)': {'ja': 'eye'},
    'Genes Immun': {'ja': 'gene'},
    'Genet Med': {'ja': 'gim'},
    'Hypertens Res': {'ja': 'hr'},
    'Int J Obes Relat Metab Disord': {'ja': 'ijo'},
    'Int J Obes (Lond)': {'ja': 'ijo'},
    'J Invest Dermatol': {'ja': 'jid'},
    'J Hum Genet': {'ja': 'jhg'},
    'J Hum Hypertens': {'ja': 'jhh'},
    'Kidney Int': {'ja': 'ki'},
    'Leukemia': {'ja': 'leu'},
    'Mod Pathol': {'ja': 'modpathol'},
    'Mol Psychiatry': {'ja': 'mp'},
    'Nature': {'ja': 'nature'},
    'Nat Biotechnol': {'ja': 'nbt'},
    'Nat Cell Biol': {'ja': 'ncb'},
    'Nat Chem': {'ja': 'nchem'},
    'Nat Clin Pract Endocrinol Metab': {'ja': 'nrendo'},
    'Nat Clin Pract Cardiovasc Med': {'ja': 'nrcardio'},
    'Nat Clin Pract Oncol': {'ja': 'nrclinonc'},
    'Nat Clin Pract Gastroenterol Hepatol': {'ja': 'nrgastro'},
    'Nat Clin Pract Urol': {'ja': 'nrurol'},
    'Nat Clin Pract Neurol': {'ja': 'nrneurol'},
    'Nat Clin Pract Nephrol': {'ja': 'nrneph'},
    'Nat Clin Pract Rheumatol': {'ja': 'nrrheum'},
    'Nat Genet': {'ja': 'ng'},
    'Nat Commun': {'ja': 'ncomms'},
    'Nat Nanotechnol': {'ja': 'nnano'},
    'Nat Neurosci': {'ja': 'neuro'},
    'Nat Mater': {'ja': 'nmat'},
    'Nat Med': {'ja': 'nm'},
    'Nat Methods': {'ja': 'nmeth'},
    'Nat Protoc': {'ja': 'nprot'},
    'Nat Rev Drug Discov': {'ja': 'nrd'},
    'Nat Rev Cardiol': {'ja': 'nrcardio'},
    'Nat Rev Clin Oncol': {'ja': 'nrclinonc'},
    'Nat Rev Endocrinol': {'ja': 'nrendo'},
    'Nat Rev Genet': {'ja': 'nrg'},
    'Nat Rev Gastroenterol Hepatol': {'ja': 'nrgastro'},
    'Nat Rev Mol Cell Biol': {'ja': 'nrm'},
    'Nat Rev Nephrol': {'ja': 'nrneph'},
    'Nat Rev Neurol': {'ja': 'nrneurol'},
    'Nat Rev Rheumatol': {'ja': 'nrrheum'},
    'Nat Rev Urol': {'ja': 'nrurol'},
    'Nat Rev Immunol': {'ja': 'nri'},
    'Neuropsychopharmacology': {'ja': 'npp'},
    'Oncogene': {'ja': 'onc'},
    'Pediatr Res': {'ja': 'pr'},
    'Pharmacogenomics J': {'ja': 'tpj'},
}



from __future__ import absolute_import

import os
import logging

PKGNAME = 'metapub'

# where to place XML (temporarily) when downloaded.
TMPDIR = '/tmp'

# default cache directory for SQLite cache engines
DEFAULT_CACHE_DIR = os.path.join(os.path.expanduser('~'), '.cache')
if not os.access(DEFAULT_CACHE_DIR, os.W_OK):
    # default cache dir is not writeable
    # therefore ask the os for a temp directory
    import tempfile
    DEFAULT_CACHE_DIR = tempfile.gettempdir()

# email address submitted to eutils with requests (as required by their api).
DEFAULT_EMAIL = os.getenv('EUTILS_EMAIL', 'metapub@nthmost.com')


def get_process_log(filepath, loglevel=logging.INFO, name=PKGNAME+'-process'):
    log = logging.getLogger(name)
    log.setLevel(loglevel)
    formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
    fh = logging.FileHandler(filepath)
    fh.setFormatter(formatter)
    fh.setLevel(loglevel)
    log.addHandler(fh)
    return log


def get_data_log(filepath, name=PKGNAME+'-data'):
    datalog = logging.getLogger(name)
    datalog.setLevel(logging.DEBUG)
    datalog.propagate = False
    formatter = logging.Formatter('')
    fh = logging.FileHandler(filepath)
    fh.setFormatter(formatter)
    datalog.addHandler(fh)
    return datalog



from __future__ import absolute_import, print_function

import unittest

from metapub import PubMedFetcher
from metapub.pubmedcentral import *


class TestPubmedFetcher(unittest.TestCase):

    def setUp(self):
        self.fetch = PubMedFetcher()

    def tearDown(self):
        pass

    def test_pmids_for_query(self):
        params = {'journal': 'PLoS One',
                  'year': 2013,
                  'author': 'McMurry AJ'}

        pmids = self.fetch.pmids_for_query(**params)
        assert len(pmids) == 1
        assert pmids[0] == '23533569'

        # this pubmed ID was deleted
        params = {'TA': 'Journal of Neural Transmission',
                  'pdat': 2014,
                  'vol': 121,
                  'aulast': 'Freitag'
                  }

        pmids = self.fetch.pmids_for_query(**params)
        assert len(pmids) == 0

    def test_medical_genetics_query(self):
        # we presume that the results for a fixed year prior to this one will not change.
        results = self.fetch.pmids_for_medical_genetics_query('Brugada Syndrome', 'diagnosis', debug=True, year=2013)
        assert '24775617' in results

    def test_clinical_query(self):
        # we presume that the results for a fixed year prior to this one will not change.
        results = self.fetch.pmids_for_clinical_query('Global developmental delay', 'etiology', 'narrow', debug=True, year=2013)
        assert results[0] == '24257216'
        assert results[1] == '24123848'
        assert results[2] == '24089199'

    def test_specified_return_slice(self):
        pmids = self.fetch.pmids_for_query(since='2015/3/1', retmax=1000)
        assert len(pmids) == 1000

        pmids = self.fetch.pmids_for_query(since='2015/3/1', retstart=200, retmax=500)
        assert len(pmids) == 500

    def test_pmc_only(self):
        params = {'mesh': 'breast neoplasm'}
        stuff = self.fetch.pmids_for_query(since='2015/1/1', until='2015/3/1', pmc_only=True, **params)
        print(stuff)



from __future__ import absolute_import, unicode_literals, print_function

import time
import logging

import six

from ..pubmedcentral import get_pmid_for_otherid
from ..pubmedfetcher import PubMedFetcher
from ..crossref import CrossRef
from ..eutils_common import SQLiteCache, get_cache_path
from ..dx_doi import DxDOI
from ..convert import doi2pmid, pmid2doi, interpret_pmids_for_citation_results
from ..exceptions import MetaPubError, DxDOIError, BadDOI
from ..utils import hostname_of, remove_chars, asciify
from ..cache_utils import datetime_to_timestamp
from ..text_mining import find_doi_in_string
from ..config import DEFAULT_CACHE_DIR

from .methods import re_pmcid, try_pmid_methods, try_doi_methods, try_vip_methods


# UrlReverse cacheing engine globals
URLREVERSE_CACHE = None
CACHE_FILENAME = 'urlreverse-cache.db'


def get_article_info_from_url(url):
    """ Using regular expressions, attempt to determine the "format" of the submitted URL, and if 
    possible, extract useful information from the URL for article lookup by ID or citation.

    Possible results:
        'vip': volume-issue-page --> {'format': 'vip', 'volume': <V>, 'issue': <I>, 'first_page': <P>, 'jtitle': <jrnl>}
        'doi': has doi in the url --> {'format': 'doi', 'doi': <DOI>, 'method': <get_doi_function>}
        'pmid': has pmid in the url --> {'format': 'pmid', 'pmid': <PMID>}
        'pmcid': has PMC id in the url --> {'format': 'pmcid': 'pmcid': <PMCID>}

    If none of the available methods work to parse the URL, the result dictionary will be:
        {'format': 'unknown'}

    :param url:
    :return: result dictionary (see above)
    """
    # maybe the DOI is deducible from the URL:
    doidict = try_doi_methods(url)
    if doidict:
        doidict['format'] = 'doi'
        return doidict

    # maybe the pubmed ID is in the URL:
    pmid = try_pmid_methods(url)
    if pmid:
        outd = {'pmid': pmid, 'format': 'pmid'}
        return outd

    # maybe the PubmedCentral ID is in the URL:
    #if 'nih.gov' in url or 'europepmc.org' in url:
    match = re_pmcid.match(url)
    if match:
        outd = match.groupdict()
        outd['format'] = 'pmcid'
        return outd

    # maybe this is a volume-issue-page formatted link and we can look it up by citation or CrossRef:
    vipdict = try_vip_methods(url)
    if vipdict:
        vipdict['format'] = 'vip'
        return vipdict

    return {'format': 'unknown'}


def _get_urlreverse_cache(cachedir=DEFAULT_CACHE_DIR):
    global URLREVERSE_CACHE
    if not URLREVERSE_CACHE:
        _cache_path = get_cache_path(cachedir, CACHE_FILENAME)
        URLREVERSE_CACHE = SQLiteCache(_cache_path)
    return URLREVERSE_CACHE


class UrlReverse(object):

    """ UrlReverse takes a url and performs the switchboard operations that hopefully lead
    to the successful "reversal" of an article url into its origination DOI and/or PMID.

    Whether the object is able to discover either or both of these identifiers depends 
    highly on the information available in the URL and inferable from what is known about
    the publisher or website that the article was found upon.

    Example:

        urlrev = UrlReverse('http://jmg.bmj.com/content/43/2/97.full.pdf')
        print(urlrev.doi)       # 10.1136/jmg.2005.030833
        print(urlrev.pmid)      # 15879500

    Human inspection can quickly verify that the above PDF definitely maps to this 
    PubMed entry:

        https://www.ncbi.nlm.nih.gov/pubmed/15879500

    (Adding a machine-verification step might be a further development of UrlReverse;
    however, it would add significant page-loading and processing time. Might be better
    off as an external "wrapper" around the UrlReverse operations.)

    The "steps" attribute will be of most interest if you want to know how UrlReverse
    arrived at its ID conclusions. 

    In the case of the above BMJ article URL, while the URL might have typically been
    "reversible" to a DOI from its constituent information, using DxDOI to verify whether
    the resultant DOI -- "10.1136/bmj.43.2.97" -- was a real one resulted in a DxDOIError,
    indicating that we did not have the Real McCoy.

    Using print(urlrev.steps), we get the following:
            
        [u'FOUND PMID via PubmedFetcher.pmids_for_citation',
         u'FOUND DOI via pmid2doi',
         u'VERIFY dx.doi.org: http://jmg.bmj.com/content/43/2/97']

    So, UrlReverse had to use a fallback method -- the pmids_for_citation approach, a 
    relatively slower method, but which in this case got the job done. This approach
    relies on the use of knowing a volume, first_page, and journal name, and 
    (hopefully) receiving a single unambiguous result from the query.

    When ambiguous results are received, UrlReverse considers this a failure (see `steps`).

    Args:

        skip_cache: (default: False) whether to load results afresh, regardless of cache contents.

    Keyword args:

        expiry_date: (default: None) forces cache to reload results older than given date.
        cachedir: (default: ~/.cache) allows change of cachedir; set to None to disable cache.
        debug: (default: False) raises log level of 'metapub.UrlReverse' logger to logging.DEBUG
    """

    def __init__(self, url, skip_cache=False, fetcher=None, dxdoi=None, crossref=None, **kwargs):
        if not url.lower().startswith('http'):
            url = 'http://' + url

        self.fetcher = fetcher or PubMedFetcher()
        self.dxdoi = dxdoi or DxDOI()
        self.cr = crossref or CrossRef()

        self.url = url
        self.steps = []

        self.pmid = None
        self.doi = None
        self.info = None

        self.expiry_date = kwargs.get('expiry_date', None)

        cachedir = kwargs.get('cachedir', DEFAULT_CACHE_DIR)
        self._cache = None if cachedir is None else _get_urlreverse_cache(cachedir)

        self._log = logging.getLogger('metapub.UrlReverse')
        if kwargs.get('debug', False):
            self._log.setLevel(logging.DEBUG)
        else:
            self._log.setLevel(logging.INFO)

        if self._cache:
            self._load_from_cache()
        else:
            self._urlreverse()

    def _urlreverse(self):
        """ the switchboard operator of the urlreverse methods.

        mutates:
            self.info
            self.format
            self.pmid
            self.doi
            self.steps
        """
        self.info = get_article_info_from_url(self.url)
        self.format = self.info['format']

        if self.format == 'pmid':
            self.pmid = self.info['pmid']
            #self.doi = pmid2doi(self.pmid)
            if self.pmid:
                self.steps.append('FOUND PMID from inferred PMID in URL')

        elif self.format == 'doi':
            self.doi = self.info['doi']
            self.steps.append('FOUND DOI via inferred doi')
            self.pmid = doi2pmid(self.doi)
            if self.pmid:
                try:
                    int(self.pmid)
                    self.steps.append('FOUND PMID via doi2pmid')
                except:
                    # we'll log this further down (avoiding repeated code).
                    pass
            else:
                self.steps.append('NO PMID from doi2pmid')

        elif self.format == 'vip':
            try:
                self._try_citation_methods()
            except MetaPubError as error:
                self.pmid = None
                self.steps.append('NO PMID from VIP info + citation methods')

        elif self.format == 'pmcid':
            self.pmid = get_pmid_for_otherid(self.info['pmcid'])
            self.doi = doi2pmid(self.pmid)
            if self.pmid:
                self.steps.append('FOUND PMID from PMCID -> PMID lookup')

        if self.pmid and self.pmid.startswith('NOT_FOUND'):
            self.steps.append('NO PMID: PMID citation lookup resulted in "%s"' % self.pmid)
            self.pmid = None

        if self.doi and not self.pmid:
            self._try_backup_doi2pmid_methods()

        if self.doi:
            try:
                urlres = self.dxdoi.resolve(self.doi)
                self.steps.append('VERIFY dx.doi.org: %s' % urlres)
            except (DxDOIError, BadDOI) as error:
                self.doi = None
                self.steps.append('VERIFY dx.doi.org: PROBLEM with DOI: %r' % error)

        # Finally: ADMIT DEFEAT
        if not self.doi and not self.pmid:
            self.steps.append('NO DOI. NO PMID. All methods failed. END OF LINE.')

    def _store_cache(self):
        """ Store this object in cache by explicitly choosing variables to store as
        values, using self.url as the cache key.

        A time.time() timestamp will be added to the value dictionary when stored.

        There is no return from this function. Exceptions from the SQLiteCache 
        object may be raised.
        """
        cache_value = self.to_dict()
        cache_value['timestamp'] = time.time()
        self._cache[self._make_cache_key(self.url)] = cache_value

    def _load_from_cache(self, retry=False, expiry_date=None):
        cache_result = self._query_cache(self.url, expiry_date)

        if cache_result:
            self.pmid = cache_result['pmid']
            self.doi = cache_result['doi']
            self.steps = cache_result['steps']
            self.info = cache_result['info']

            if retry:
                if 'END OF LINE' in ';'.join(self.steps):
                    self._urlreverse()
                    self._store_cache()

        else:
            self._urlreverse()
            self._store_cache()

    def _make_cache_key(self, url):
        """ Returns url normalized via str() function for hash lookup / store. """
        return str(url)

    def _query_cache(self, cache_key, expiry_date=None):
        """ Return results of a lookup from the cache, if available.
        Return None if not available.

        Cache results are stored with a time.time() timestamp.

        When expiry_date is supplied, results from the cache past their
        sell-by date will be expunged from the cache and return will be None.

        expiry_date can be either a python datetime or a timestamp. 

        :param: cache_key: (required)
        :param: expiry_date (optional, default None)
        :return: (dict) result of cache lookup
        :rtype: dict or None
        """

        if hasattr(expiry_date, 'strftime'):
            # convert to timestamp
            sellby = datetime_to_timestamp(expiry_date)
        else:
            # make sure sellby is a number, not None
            sellby = expiry_date if expiry_date else 0

        if self._cache:
            cache_key = self._make_cache_key(cache_key)
            try:
                res = self._cache[cache_key]
                timestamp = res['timestamp']
                if timestamp < sellby:
                    self._log.debug('Cache: expunging result for %s (%i)', cache_key, timestamp)
                else:
                    self._log.debug('Cache: returning result for %s (%i)', cache_key, timestamp)
                return res

            except KeyError:
                self._log.debug('Cache: no result for key %s', cache_key)
                return None
        else:
            self._log.debug('Cache disabled (self._cache is None)')
            return None

    def _try_citation_methods(self):
        # 1) try pubmed citation match.
        pmids = self.fetcher.pmids_for_citation(**self.info)
        pmid = interpret_pmids_for_citation_results(pmids)
        if pmid and pmid != 'AMBIGUOUS':
            self.pmid = pmid
            self.doi = pmid2doi(pmid)
            self.steps.append('FOUND PMID via PubmedFetcher.pmids_for_citation')
            if self.doi:
                self.steps.append('FOUND DOI via pmid2doi')
            return

        # 2) try CrossRef -- most effective when title available, but may work without it.
        results = self.cr.query('', params=self.info)
        if results:
            top_result = self.cr.get_top_result(results, self.cr.last_params)

            # we may have disqualified all the results at this point as being irrelevant, so we have to test here.
            if top_result:
                self.doi = find_doi_in_string(top_result['doi'])
                pmids = self.fetcher.pmids_for_citation(**top_result['slugs'])
                pmid = interpret_pmids_for_citation_results(pmids)
                if pmid and pmid != 'AMBIGUOUS':
                    self.pmid = pmid
                else:
                    self.steps.append('NO PMID: AMBIGUOUS results from citation lookup (CrossRef result was %r)' % top_result)
            else:
                self.steps.append('NO PMID: Zero results from CrossRef while trying citation methods')

    def _try_backup_doi2pmid_methods(self):
        """ Uses CrossRef and Pubmed Advanced Query combinations to try to get an 
        unambiguous PMID result. Mutates self.pmid (if found unambigously) and self.steps
        (appending strings documenting the process by which PMID was(n't) acquired).
        """

        # All hinges on whether CrossRef can give us a good result. If not, fail out early.
        results = self.cr.query(self.doi)
        coins = None

        if results:
            top_result = self.cr.get_top_result(results, self.cr.last_params)
            if top_result:
                coins = top_result['slugs'].copy()

        if not coins:
            return

        # normalize start page ('spage') to None in case of "no" or "n/a"
        if coins.get('spage', 'no').lower() in ['no', 'n%2Fa']:
            coins['spage'] = None

        # bowlderize the title (remove urlencoded chars, unicode-only chars, and punctuation).
        # ps. some entries have no title (really!)
        try:
            if six.PY2:
                title = remove_chars(coins.get('atitle', ''), urldecode=True).strip().encode('ascii', errors='ignore')
            else:
                title = remove_chars(coins.get('atitle', ''), urldecode=True).strip()
        except KeyError:
            self.steps.append('CrossRef result has no title. This bodes not well. (CrossRef result was %r)' % top_result)

        pmids = []

        if title:
            # try just searching by title first. If we get one single result, that should be it.
            pmids = self.fetcher.pmids_for_query(title)

            if len(pmids) == 1:
                self.pmid = pmids[0]
                self.steps.append('FOUND PMID via Pubmed Advanced Query')
                return

            elif len(pmids) == 0:
                self.pmid = None
                self.steps.append('Zero results for title "%s" in Pubmed, attempting coordinate match' % title)
                title = ''

            elif len(pmids) > 1 and len(title.split(' ')) < 3:
                # title could be something like "Abstract" or "Pituitary" or "Endocrinology Yearbook" -- too vague.
                self.steps.append('Title "%s" too VAGUE, attempting coordinate match' % title)
                title = ''

        # we have ambiguous results (or no title at all) -- let's try to narrow the field based on
        # whether we have a viable title or not.

        # Two paths diverged in a wood, and I...

        if title=='':
            # strict coordinates
            params = {'VI': coins.get('volume', None),
                      'IP': coins.get('issue', None),
                      'AU': coins.get('aulast', None),
                      'PG': coins.get('spage', None),
                      'DP': coins.get('year', None),
                     }
            try:
                pmids = self.fetcher.pmids_for_query(coins['jtitle'], **params)
            except KeyError:
                # hrm, no title and no jtitle, eh... let's bail.
                self.steps.append('NO PMID. CrossRef data unworkable (no jtitle). END OF LINE.')
                return 

        else:
            if coins.get('volume') and coins.get('issue'):
                self.steps.append('AMBIGUOUS results for title "%s", trying with volume/issue')
                pmids = self.fetcher.pmids_for_query(title, VI=coins['volume'], IP=coins['issue'])
            elif coins.get('volume') and coins.get('aulast'):
                self.steps.append('AMBIGUOUS results for title "%s", trying with aulast')
                pmids = self.fetcher.pmids_for_query(title, AU=coins['aulast'])
            elif coins.get('spage') and coins.get('aulast'):
                self.steps.append('AMBIGUOUS results for title "%s", trying with first_page')
                pmids = self.fetcher.pmids_for_query(title, PG=coins['spage'])     #, AU=coins['aulast'])
            elif coins.get('volume'):
                self.steps.append('AMBIGUOUS results for title "%s", trying with volume')
                pmids = self.fetcher.pmids_for_query(title, VI=coins['volume'])

        # that should have narrowed the field substantially. we should give up if it's still ambiguous.
        if len(pmids) == 1:
            self.pmid = pmids[0]
            self.steps.append('FOUND PMID via Pubmed Advanced Query')
        elif len(pmids) == 0:
            self.pmid = None
            self.steps.append('NO PMID, zero results from pubmed advanced query. (Data from CrossRef was: %r)' % (coins))
        else:
            self.pmid = None
            self.steps.append('NO PMID, AMBIGUOUS results from pubmed advanced query (%i possibilities). (Data from CrossRef was: %r)' % (len(pmids), coins))

    def to_dict(self):
        """ Returns a dictionary containing all public object attributes (i.e. not starting with an underscore). """
        outd = {}
        for key in self.__dict__:
            if not key.startswith('_'):
                if not key in ['dxdoi', 'cr', 'fetcher']:
                    outd[key] = self.__dict__[key]
        return outd




from __future__ import absolute_import, unicode_literals

# Science (AAAS) -- requires login "as a courtesy to the community".  Mkay.
aaas_format = 'http://{ja}.sciencemag.org/content/{a.volume}/{a.issue}/{a.first_page}.full.pdf'
aaas_journals = {
    'Science': {'ja': 'www'},
    'Sci Adv': {'ja': 'advances'},
    'Sci Signal': {'ja': 'stke'},
    'Sci Transl Med': {'ja': 'stm'},
}




from __future__ import absolute_import, unicode_literals

jama_journals = (
    'AMA Arch Intern Med',
    'Arch Dermatol',
    'Arch Gen Psychiatry',
    'Arch Neurol',
    'Arch Ophthalmol',
    'Arch Surg',
    'JAMA',
    'JAMA Dermatol',
    'JAMA Facial Plast Surg',
    'JAMA Intern Med',
    'JAMA Neurol',
    'JAMA Oncol',
    'JAMA Ophthalmol',
    'JAMA Otolaryngol Head Neck Surg',
    'JAMA Pediatr',
    'JAMA Psychiatry',
    'JAMA Surg',
)



from __future__ import absolute_import, unicode_literals

"""metapub.pubmedauthor -- PubMedAuthor class instantiated a ncbi Author XML Element"""

from .base import MetaPubObject
from .exceptions import MetaPubError


class PubMedAuthor(MetaPubObject):
    """This PubMedAuthor class receives a xml element as required argument
    and parses it into its parts, exposing them as attributes.

    Usage:
        author = PubMedAuthor(xml_elem)

    To retrieve the standard represenation of a author name, use the __str__ method.

    (About unicode: metapub uses unicode_literals in both py3 and py2, so the str() function
    returns unicode, unless called by a py2k "str()" statement in which unicode_literals is off.)
    """

    def __init__(self, xmlelem, *args, **kwargs):

        self.last_name = None
        self.fore_name = None
        self.initials = None
        self.collective_name = None
        self.affiliations = []

        if xmlelem is None:
            return

        self.content = xmlelem
        self._parse_xml()

    def to_dict(self):
        outd = self.__dict__.copy()
        outd.pop(content)
        return outd

    def _parse_xml(self):
        if self.content is None:
            return

        try:
            self.last_name = self.content.find('LastName').text
        except AttributeError:
            pass

        try:
            self.fore_name = self.content.find('ForeName').text
        except AttributeError:
            pass

        try:
            self.initials = self.content.find('Initials').text
        except AttributeError:
            pass

        try:
            self.collective_name = self.content.find('CollectiveName').text
        except AttributeError:
            pass

        try:
            self.affiliations = [aff.text for aff in self.content.find('AffiliationInfo').findall('Affiliation')]
        except AttributeError:
            pass

        if self.last_name is None and self.fore_name is None and self.initials is None and self.collective_name is None and self.affiliations == []:
            raise MetaPubError('Author structure not recognized')

    def __str__(self):
        if self.last_name and self.initials:
            return self.last_name + ' ' + self.initials
        elif self.collective_name:
            return self.collective_name
        elif self.last_name:
            return self.last_name
        else:
            return 'Unnamed Author'




from __future__ import absolute_import, unicode_literals

sciencedirect_url = 'http://www.sciencedirect.com/science/article/pii/{piit}'

sciencedirect_journals = [
        'Acad Pediatr',
        'Acta Astronaut',
        'Acta Biomater',
        'Acta Cardiol',
        'Acta Histochem',
        'Acta Reumatol Port',
        'Acta Trop',
        'Actas Dermosifiliogr',
        'Actas Urol Esp',
        'Adv Biophys',
        'Adv Cancer Res',
        'Adv Clin Chem',
        'Adv Colloid Interface Sci',
        'Adv Drug Deliv Rev',
        'Adv Enzyme Regul',
        'Adv Med Sci',
        'Adv Protein Chem',
        'Adv Virus Res',
        'Ageing Res Rev',
        'Air Med J',
        'Alcohol',
        'Allergol Immunopathol (Madr)',
        'Allergol Int',
        'Alzheimers Dement',
        'Am J Emerg Med',
        'Am J Geriatr Psychiatry',
        'Am J Kidney Dis',
        'Am J Obstet Gynecol',
        'Am J Orthod Dentofacial Orthop',
        'Am J Otolaryngol',
        'Am J Pathol',
        'Am J Prev Med',
        'Am J Surg',
        'Ambul Pediatr',
        'An Pediatr (Barc)',
        'Anal Biochem',
        'Anal Chim Acta',
        'Anim Reprod Sci',
        'Ann Acad Med Singap',
        'Ann Allergy Asthma Immunol',
        'Ann Chir Gynaecol',
        'Ann Diagn Pathol',
        'Ann Emerg Med',
        'Ann Endocrinol (Paris)',
        'Ann Epidemiol',
        'Ann Genet',
        'Ann Hepatol',
        'Ann Inst Pasteur Immunol',
        'Ann Inst Pasteur Microbiol',
        'Ann Intern Med',
        'Ann Pathol',
        'Ann Thorac Surg',
        'Ann Transplant',
        'Ann Vasc Surg',
        'Antivir Ther (Lond)',
        'Antiviral Res',
        'Appetite',
        'Appl Radiat Isot',
        'Aquat Toxicol',
        'Arch Biochem Biophys',
        'Arch Bronconeumol',
        'Arch Cardiovasc Dis',
        'Arch Gerontol Geriatr',
        'Arch Gerontol Geriatr Suppl',
        'Arch Med Res',
        'Arch Oral Biol',
        'Arch Pediatr',
        'Arch Soc Esp Oftalmol',
        'Arthroscopy',
        'Asian J Psychiatr',
        'Asian J Surg',
        'Asian Pac J Trop Med',
        'Aten Primaria',
        'Atheroscler Suppl',
        'Atherosclerosis',
        'Auris Nasus Larynx',
        'Aust J Physiother',
        'Autoimmun Rev',
        'Auton Neurosci',
        'Behav Brain Res',
        'Behav Neural Biol',
        'Behav Processes',
        'Beijing Da Xue Xue Bao',
        'Best Pract Res Clin Gastroenterol',
        'Best Pract Res Clin Haematol',
        'Best Pract Res Clin Rheumatol',
        'BioSystems',
        'Biochem Biophys Res Commun',
        'Biochem Med',
        'Biochem Med Metab Biol',
        'Biochem Mol Med',
        'Biochem Pharmacol',
        'Biochim Biophys Acta',
        'Biochimie',
        'Bioelectrochem Bioenerg',
        'Bioelectrochemistry',
        'Bioinorg Chem',
        'Biol Blood Marrow Transplant',
        'Biol Psychol',
        'Biologicals',
        'Biomaterials',
        'Biomed Environ Sci',
        'Biomed Pharmacother',
        'Biomol Eng',
        'Bioorg Chem',
        'Bioorg Med Chem',
        'Bioorg Med Chem Lett',
        'Biophys Chem',
        'Bioresour Technol',
        'Biosens Bioelectron',
        'Biotechnol Adv',
        'Blood Cells Mol Dis',
        'Blood Rev',
        'Bone Miner',
        'Br J Oral Maxillofac Surg',
        'Br J Plast Surg',
        'Br Vet J',
        'Brain Behav Immun',
        'Brain Cogn',
        'Brain Res',
        'Brain Res Brain Res Protoc',
        'Brain Res Brain Res Rev',
        'Brain Res Bull',
        'Brain Res Dev Brain Res',
        'Brain Res Gene Expr Patterns',
        'Brain Res Mol Brain Res',
        'Brain Stimul',
        'Braz J Anesthesiol',
        'Braz J Infect Dis',
        'Braz J Otorhinolaryngol',
        'Breast',
        'Bull Cancer',
        'Burns',
        'C R Acad Sci III, Sci Vie',
        'C R Biol',
        'Can J Cardiol',
        'Can J Diabetes',
        'Can J Ophthalmol',
        'Cancer Detect Prev',
        'Cancer Epidemiol',
        'Cancer Genet',
        'Cancer Radiother',
        'Cancer Treat Rev',
        'Carbohydr Polym',
        'Carbohydr Res',
        'Cardiovasc Pathol',
        'Cardiovasc Surg',
        'Cas Lek Cesk',
        'Cell Calcium',
        'Cell Differ Dev',
        'Cell Immunol',
        'Cell Rep',
        'Cell Signal',
        'Chem Biol Interact',
        'Chem Phys',
        'Chem Phys Lett',
        'Chem Phys Lipids',
        'Chemosphere',
        'Child Abuse Negl',
        'Child Adolesc Psychiatr Clin N Am',
        'Chin J Nat Med',
        'Chin Med J',
        'Chin Med Sci J',
        'Clin Biochem',
        'Clin Biomech (Bristol, Avon)',
        'Clin Breast Cancer',
        'Clin Chim Acta',
        'Clin Colorectal Cancer',
        'Clin Exp Rheumatol',
        'Clin Gastroenterol Hepatol',
        'Clin Genitourin Cancer',
        'Clin Immunol',
        'Clin Immunol Immunopathol',
        'Clin Investig Arterioscler',
        'Clin Lab Med',
        'Clin Lab Sci',
        'Clin Lung Cancer',
        'Clin Lymphoma Myeloma',
        'Clin Lymphoma Myeloma Leuk',
        'Clin Microbiol Infect',
        'Clin Neurophysiol',
        'Clin Nutr',
        'Clin Oncol (R Coll Radiol)',
        'Clin Psychol Rev',
        'Clin Res Hepatol Gastroenterol',
        'Clin Ther',
        'Colloids Surf B Biointerfaces',
        'Comp Biochem Physiol, B',
        'Comp Biochem Physiol B, Biochem Mol Biol',
        'Comp Biochem Physiol C, Pharmacol Toxicol Endocrinol',
        'Comp Biochem Physiol C Toxicol Pharmacol',
        'Comp Biochem Physiol, Part A Mol Integr Physiol',
        'Comp Biochem Physiol Part D Genomics Proteomics',
        'Comp Immunol Microbiol Infect Dis',
        'Compr Psychiatry',
        'Comput Biol Chem',
        'Comput Biol Med',
        'Comput Chem',
        'Comput Methods Programs Biomed',
        'Comput Radiol',
        'Cont Lens Anterior Eye',
        'Contemp Clin Trials',
        'Contraception',
        'Cortex',
        'Crit Care Clin',
        'Crit Rev Oncol Hematol',
        'Curr Opin Biotechnol',
        'Curr Opin Chem Biol',
        'Curr Opin Gastroenterol',
        'Curr Opin Genet Dev',
        'Curr Opin Plant Biol',
        'Curr Opin Struct Biol',
        'Curr Opin Virol',
        'Curr Probl Pediatr',
        'Cytokine',
        'Cytokine Growth Factor Rev',
        'Cytotherapy',
        'DNA Repair (Amst)',
        'Dent Mater',
        'Dev Biol',
        'Dev Comp Immunol',
        'Development',
        'Diabetes Metab',
        'Diabetes Metab Syndr',
        'Diagn Microbiol Infect Dis',
        'Dig Liver Dis',
        'Domest Anim Endocrinol',
        'Drug Alcohol Depend',
        'Drug Resist Updat',
        'Early Hum Dev',
        'Ecotoxicol Environ Saf',
        'Electroencephalogr Clin Neurophysiol',
        'Encephale',
        'Endocrinol Metab Clin North Am',
        'Endocrinol Nutr',
        'Enferm Infecc Microbiol Clin',
        'Environ Int',
        'Environ Pollut',
        'Environ Res',
        'Environ Toxicol Pharmacol',
        'Enzyme Microb Technol',
        'Epilepsy Behav',
        'Eur Ann Otorhinolaryngol Head Neck Dis',
        'Eur J Cancer',
        'Eur J Cancer Clin Oncol',
        'Eur J Cell Biol',
        'Eur J Gynaecol Oncol',
        'Eur J Intern Med',
        'Eur J Med Chem',
        'Eur J Med Genet',
        'Eur J Obstet Gynecol Reprod Biol',
        'Eur J Pharm Biopharm',
        'Eur J Pharm Sci',
        'Eur J Pharmacol',
        'Eur J Radiol',
        'Eur J Surg Oncol',
        'Eur J Ultrasound',
        'Eur J Vasc Endovasc Surg',
        'Eur Neuropsychopharmacol',
        'Eur Psychiatry',
        'Eur Urol',
        'Exp Cell Res',
        'Exp Eye Res',
        'Exp Gerontol',
        'Exp Mol Pathol',
        'Exp Neurol',
        'Exp Parasitol',
        'Exp Pathol (Jena)',
        'Exp Toxicol Pathol',
        'FEBS Lett',
        'Fam Pract',
        'Farm Hosp',
        'Farmaco',
        'Fish Shellfish Immunol',
        'Fitoterapia',
        'Fold Des',
        'Food Chem',
        'Food Chem Toxicol',
        'Food Microbiol',
        'Forensic Sci Int',
        'Forensic Sci Int Genet',
        'Free Radic Biol Med',
        'Fundam Appl Toxicol',
        'Fungal Biol',
        'Fungal Genet Biol',
        'Gac Sanit',
        'Gait Posture',
        'Gastroenterol Clin Biol',
        'Gastroenterol Clin North Am',
        'Gastroenterol Hepatol',
        'Gastrointest Endosc',
        'Gen Comp Endocrinol',
        'Gen Pharmacol',
        'Gend Med',
        'Gene',
        'Gene Expr Patterns',
        'Genet Anal',
        'Genet Anal Tech Appl',
        'Genomics',
        'Genomics Proteomics Bioinformatics',
        'Growth Horm IGF Res',
        'Gynecol Obstet Fertil',
        'HBPD INT',
        'Handb Clin Neurol',
        'Health Policy',
        'Hear Res',
        'Heart Lung',
        'Heart Lung Circ',
        'Hematol Oncol Clin North Am',
        'Hematol Oncol Stem Cell Ther',
        'Hepatol Res',
        'Homeopathy',
        'Homo',
        'Horm Behav',
        'Hum Immunol',
        'Hum Pathol',
        'Icarus',
        'Immunobiology',
        'Immunochemistry',
        'Immunol Lett',
        'Immunol Today',
        'Immunomethods',
        'Immunopharmacology',
        'Immunotechnology',
        'Infect Genet Evol',
        'Injury',
        'Insect Biochem Mol Biol',
        'Int Angiol',
        'Int Immunopharmacol',
        'Int J Antimicrob Agents',
        'Int J Biochem',
        'Int J Biochem Cell Biol',
        'Int J Biol Macromol',
        'Int J Dev Neurosci',
        'Int J Food Microbiol',
        'Int J Gynaecol Obstet',
        'Int J Hyg Environ Health',
        'Int J Immunopharmacol',
        'Int J Infect Dis',
        'Int J Med Microbiol',
        'Int J Obstet Anesth',
        'Int J Oral Maxillofac Surg',
        'Int J Parasitol',
        'Int J Pediatr Otorhinolaryngol',
        'Int J Pharm',
        'Int J Psychophysiol',
        'Int J Rad Appl Instrum A',
        'Int J Rad Appl Instrum B',
        'Int J Radiat Oncol Biol Phys',
        'Int Rev Cytol',
        'Int Rev Neurobiol',
        'J AAPOS',
        'J Acad Nutr Diet',
        'J Acupunct Meridian Stud',
        'J Adolesc',
        'J Adolesc Health',
        'J Affect Disord',
        'J Allergy Clin Immunol',
        'J Allergy Clin Immunol Pract',
        'J Am Acad Child Adolesc Psychiatry',
        'J Am Acad Dermatol',
        'J Am Coll Cardiol',
        'J Am Coll Surg',
        'J Am Dent Assoc',
        'J Am Diet Assoc',
        'J Am Soc Echocardiogr',
        'J Antibiot',
        'J Arthroplasty',
        'J Autoimmun',
        'J Auton Nerv Syst',
        'J Biochem Biophys Methods',
        'J Biomech',
        'J Biosci Bioeng',
        'J Biotechnol',
        'J Card Fail',
        'J Cardiol',
        'J Cardiothorac Vasc Anesth',
        'J Cataract Refract Surg',
        'J Chem Neuroanat',
        'J Chin Med Assoc',
        'J Chromatogr',
        'J Chromatogr A',
        'J Chromatogr B Analyt Technol Biomed Life Sci',
        'J Chromatogr B, Biomed Appl',
        'J Chromatogr B Biomed Sci Appl',
        'J Clin Anesth',
        'J Clin Densitom',
        'J Clin Epidemiol',
        'J Clin Gastroenterol',
        'J Clin Lipidol',
        'J Clin Neurosci',
        'J Clin Psychopharmacol',
        'J Clin Virol',
        'J Colloid Interface Sci',
        'J Comp Pathol',
        'J Contam Hydrol',
        'J Control Release',
        'J Craniomaxillofac Surg',
        'J Cyst Fibros',
        'J Dairy Res',
        'J Dairy Sci',
        'J Dent',
        'J Dermatol Sci',
        'J Diabetes Complicat',
        'J Egypt Natl Canc Inst',
        'J Electrocardiol',
        'J Emerg Med',
        'J Endod',
        'J Environ Manage',
        'J Environ Sci (China)',
        'J Ethnopharmacol',
        'J Foot Ankle Surg',
        'J Formos Med Assoc',
        'J Fr Ophtalmol',
        'J Genet Genomics',
        'J Gynecol Obstet Biol Reprod (Paris)',
        'J Hand Ther',
        'J Hazard Mater',
        'J Health Econ',
        'J Heart Lung Transplant',
        'J Hepatol',
        'J Hosp Infect',
        'J Hum Evol',
        'J Immunol Methods',
        'J Infect',
        'J Infect Chemother',
        'J Inorg Biochem',
        'J Insect Physiol',
        'J Invertebr Pathol',
        'J Invest Dermatol',
        'J Lab Clin Med',
        'J Magn Reson',
        'J Magn Reson B',
        'J Mal Vasc',
        'J Mech Behav Biomed Mater',
        'J Mechanochem Cell Motil',
        'J Med Dent Sci',
        'J Microbiol Immunol Infect',
        'J Microbiol Methods',
        'J Minim Invasive Gynecol',
        'J Mol Biol',
        'J Mol Graph Model',
        'J Mol Spectrosc',
        'J Mycol Med',
        'J Neuroimmunol',
        'J Neurol Sci',
        'J Neurosci Methods',
        'J Nurse Midwifery',
        'J Nutr Biochem',
        'J Oral Maxillofac Surg',
        'J Pain',
        'J Pediatr',
        'J Pediatr Health Care',
        'J Pediatr Surg',
        'J Pharm Biomed Anal',
        'J Pharmacol Sci',
        'J Pharmacol Toxicol Methods',
        'J Photochem Photobiol B, Biol',
        'J Plant Physiol',
        'J Prosthet Dent',
        'J Prosthodont Res',
        'J Protein Chem',
        'J Proteomics',
        'J Psychiatr Res',
        'J Psychosom Res',
        'J Radiat Res',
        'J Ren Nutr',
        'J Reprod Immunol',
        'J Sci Med Sport',
        'J Shoulder Elbow Surg',
        'J Soc Gynecol Investig',
        'J Steroid Biochem',
        'J Steroid Biochem Mol Biol',
        'J Stroke Cerebrovasc Dis',
        'J Struct Biol',
        'J Surg Res',
        'J Theor Biol',
        'J Thorac Cardiovasc Surg',
        'J Trace Elem Med Biol',
        'J Tradit Chin Med',
        'J Urol',
        'J Vasc Surg',
        'J Vet Cardiol',
        'J Virol Methods',
        'JACC Cardiovasc Imaging',
        'JACC Cardiovasc Interv',
        'Joint Bone Spine',
        'Jpn J Ophthalmol',
        'Kaohsiung J Med Sci',
        'Klin Oczna',
        'Knee',
        'Lab Invest',
        'Leg Med (Tokyo)',
        'Leuk Res',
        'Life Sci',
        'Liver Transpl Surg',
        'Lung Cancer',
        'MAGMA',
        'Magn Reson Imaging',
        'Mar Environ Res',
        'Mar Pollut Bull',
        'Mater Sci Eng C Mater Biol Appl',
        'Math Biosci',
        'Matrix',
        'Matrix Biol',
        'Maturitas',
        'Mayo Clin Proc',
        'Meat Sci',
        'Mech Ageing Dev',
        'Mech Dev',
        'Med Dosim',
        'Med Mal Infect',
        'Medicina (Kaunas)',
        'Metab Eng',
        'Meth Enzymol',
        'Methods',
        'Methods Mol Med',
        'Microb Pathog',
        'Microbes Infect',
        'Microbiol Res',
        'Microvasc Res',
        'Mikrobiyol Bul',
        'Mitochondrion',
        'Mol Aspects Med',
        'Mol Biochem Parasitol',
        'Mol Cell Biol Res Commun',
        'Mol Cell Endocrinol',
        'Mol Cell Neurosci',
        'Mol Cell Probes',
        'Mol Diagn',
        'Mol Gen Genet',
        'Mol Genet Metab',
        'Mol Immunol',
        'Mol Oncol',
        'Mol Phylogenet Evol',
        'Mol Plant',
        'Mult Scler Relat Disord',
        'Mutat Res',
        'Mutat Res Genet Toxicol Environ Mutagen',
        'Mutat Res Rev Mutat Res',
        'Mycol Res',
        'N Biotechnol',
        'Nanomedicine',
        'Nephrol Ther',
        'Neth J Med',
        'Neural Netw',
        'Neurobiol Dis',
        'Neurobiol Learn Mem',
        'Neurochem Int',
        'Neurochirurgie',
        'Neuroimage',
        'Neurol Neurochir Pol',
        'Neurologia',
        'Neuromuscul Disord',
        'Neuropeptides',
        'Neuropharmacology',
        'Neuropsychologia',
        'Neurosci Biobehav Rev',
        'Neurosci Lett',
        'Neurosci Res',
        'Neuroscience',
        'Neurotoxicol Teratol',
        'Neurotoxicology',
        'Nippon Rinsho',
        'Nitric Oxide',
        'Nucl Med Biol',
        'Nucleic Acids Res Suppl',
        'Nutr Hosp',
        'Nutr Metab Cardiovasc Dis',
        'Nutr Res',
        'Nutrition',
        'Obes Res Clin Pract',
        'Oral Oncol',
        'Oral Surg Oral Med Oral Pathol',
        'Oral Surg Oral Med Oral Pathol Oral Radiol',
        'Oral Surg Oral Med Oral Pathol Oral Radiol Endod',
        'Orthop Traumatol Surg Res',
        'Osteoarthr Cartil',
        'Otolaryngol Pol',
        'Paediatr Respir Rev',
        'Pancreas',
        'Pancreatology',
        'Parasitol Int',
        'Parkinsonism Relat Disord',
        'Pathol Biol',
        'Pathol Res Pract',
        'Pathophysiology',
        'Patient Educ Couns',
        'Pediatr Neonatol',
        'Pediatr Nurs',
        'Pediatr Pulmonol',
        'Peptides',
        'Pestic Biochem Physiol',
        'Pharmacol Biochem Behav',
        'Pharmacol Res',
        'Pharmacol Ther',
        'Photodiagnosis Photodyn Ther',
        'Phys Med',
        'Phys Rev, D',
        'Physiol Behav',
        'Phytochemistry',
        'Phytomedicine',
        'Plant Physiol Biochem',
        'Plant Sci',
        'Plasmid',
        'Postepy Hig Med Dosw (Online)',
        'Pract Radiat Oncol',
        'Pregnancy Hypertens',
        'Presse Med',
        'Prev Med',
        'Prev Vet Med',
        'Prog Biophys Mol Biol',
        'Prog Brain Res',
        'Prog Neuropsychopharmacol Biol Psychiatry',
        'Prog Nucleic Acid Res Mol Biol',
        'Prog Orthod',
        'Prostaglandins',
        'Prostaglandins Leukot Essent Fatty Acids',
        'Prostaglandins Leukot Med',
        'Prostaglandins Other Lipid Mediat',
        'Protein Expr Purif',
        'Protist',
        'Prz Lek',
        'Psychiatry Res',
        'Psychoneuroendocrinology',
        'Pulm Pharmacol',
        'Radiat Meas',
        'Radiat Med',
        'Radiother Oncol',
        'Regul Pept',
        'Regul Toxicol Pharmacol',
        'Reprod Biol',
        'Reprod Biomed Online',
        'Reprod Toxicol',
        'Res Dev Disabil',
        'Res Microbiol',
        'Res Vet Sci',
        'Res Virol',
        'Respir Investig',
        'Respir Med',
        'Respir Physiol',
        'Respir Physiol Neurobiol',
        'Resuscitation',
        'Reumatol Clin',
        'Rev Argent Microbiol',
        'Rev Bras Anestesiol',
        'Rev Bras Reumatol',
        'Rev Clin Esp',
        'Rev Epidemiol Sante Publique',
        'Rev Esp Cardiol (Engl Ed)',
        'Rev Iberoam Micol',
        'Rev Mal Respir',
        'Rev Med Inst Mex Seguro Soc',
        'Rev Med Interne',
        'Rev Neurol',
        'Rev Neurol (Paris)',
        'Rev Pneumol Clin',
        'Rev Psiquiatr Salud Ment',
        'Ross Fiziol Zh Im I M Sechenova',
        'Sci Total Environ',
        'Seizure',
        'Semin Arthritis Rheum',
        'Semin Cancer Biol',
        'Semin Cell Biol',
        'Semin Cell Dev Biol',
        'Semin Diagn Pathol',
        'Semin Fetal Neonatal Med',
        'Semin Hematol',
        'Semin Nephrol',
        'Semin Oncol',
        'Semin Pediatr Neurol',
        'Semin Perinatol',
        'Semin Vasc Surg',
        'Sheng Wu Gong Cheng Xue Bao',
        'Sleep Med',
        'Sleep Med Rev',
        'Soc Sci Med',
        'Solid State Nucl Magn Reson',
        'Spectrochim Acta A Mol Biomol Spectrosc',
        'Spine J',
        'Steroids',
        'Suppl Clin Neurophysiol',
        'Surg Obes Relat Dis',
        'Surg Oncol',
        'Surgery',
        'Surv Ophthalmol',
        'Syst Appl Microbiol',
        'Taiwan J Obstet Gynecol',
        'Talanta',
        'Tetrahedron Lett',
        'Theriogenology',
        'Toxicol Appl Pharmacol',
        'Toxicol In Vitro',
        'Toxicol Lett',
        'Toxicol Sci',
        'Toxicology',
        'Toxicon',
        'Transfus Apher Sci',
        'Transfus Clin Biol',
        'Transl Res',
        'Transpl Immunol',
        'Transplant Proc',
        'Travel Med Infect Dis',
        'Trends Biochem Sci',
        'Trends Cardiovasc Med',
        'Trends Endocrinol Metab',
        'Trends Genet',
        'Trends Immunol',
        'Trends Microbiol',
        'Trends Neurosci',
        'Trends Parasitol',
        'Tuber Lung Dis',
        'Tuberculosis (Edinb)',
        'Ultramicroscopy',
        'Ultrason Sonochem',
        'Ultrasonics',
        'Ultrasound Med Biol',
        'Urol Oncol',
        'Urology',
        'Vaccine',
        'Vasc Med',
        'Vascul Pharmacol',
        'Vet Immunol Immunopathol',
        'Vet J',
        'Vet Microbiol',
        'Vet Parasitol',
        'Virology',
        'Virus Res',
        'Vision Res',
        'Vitam Horm',
        'Waste Manag',
        'Water Res',
        'Wei Sheng Yan Jiu',
        'Wilderness Environ Med',
        'Yi Chuan',
        'Yi Chuan Xue Bao',
        'Zentralbl Bakteriol',
        'Zh Nevrol Psikhiatr Im S S Korsakova',
        'Zhonghua Fu Chan Ke Za Zhi',
        'Zhonghua Yan Ke Za Zhi',
        'Zhonghua Yi Xue Za Zhi',
]
